import vdomr as vd
import time
import sys
import mtlogging
import numpy as np
from copy import deepcopy
from spikeforest import EfficientAccessRecordingExtractor
from numpy.linalg import svd


class ClusterView(vd.Component):
    def __init__(self, *, context, opts=None, prepare_result=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._features = prepare_result['features']
        self._labels = prepare_result['labels']
        self._feature_names = prepare_result['feature_names']
        self._widget = ClusterWidget(features=self._features, labels=self._labels, feature_names=self._feature_names)
        self._widget.setSize(self._size)
        self.refresh()

    @staticmethod
    def prepareView(context, opts):
        sorting_context = context
        recording_context = context.recordingContext()
        sorting_context.initialize()

        sorting = sorting_context.sortingExtractor()

        unit_ids = sorting_context.selectedUnitIds()
        earx = EfficientAccessRecordingExtractor(recording=recording_context.recordingExtractor())

        M = earx.get_num_channels()
        T = 50
        # K = len(unit_ids)
        num_features = 2

        spikes_list = []
        for unit_id in unit_ids:
            spikes = _get_random_spike_waveforms(sorting=sorting, recording=earx, unit=unit_id, max_num=50, channels=None, snippet_len=T)
            spikes_list.append(spikes)

        feature_waveforms = _get_feature_waveforms_for_spikes(spikes_list=spikes_list, num_features=num_features)
        # (num_features) x M x T
        features_list = []
        labels_list = []
        for ii, unit_id in enumerate(unit_ids):
            spikes0 = spikes_list[ii]
            features0 = np.reshape(feature_waveforms, (num_features, M * T)) @ np.reshape(spikes0, (M * T, spikes0.shape[2]))
            features_list.append(features0)
            labels_list.append(np.ones((features0.shape[1])) * unit_id)

        features = np.concatenate(features_list, axis=1)
        labels = np.concatenate(labels_list)
        return dict(
            features=features,
            labels=labels,
            feature_names=['f{}'.format(i) for i in range(features.shape[0])]
        )

    def setSize(self, size):
        if self._size != size:
            self._size = size
        if self._widget:
            self._widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Clusters'

    def title(self):
        return 'Clusters for {}'.format(self._context.sortingLabel())

    def render(self):
        return self._widget


class ClusterWidget(vd.Component):
    def __init__(self, *, features, labels, feature_names):
        vd.Component.__init__(self)
        self._features = features
        self._labels = labels
        self._feature_names = feature_names
        self._size = (800, 500)
        self._marker_size = 14
        self._hover_marker_size = 18
        self._plot = None

    def setSize(self, size):
        if self._size == size:
            return
        self._size = size
        if self._plot:
            self._plot.updateSize(size)

    def render(self):
        unit_ids = sorted(list(set(self._labels.tolist())))
        data = []
        for unit_id in unit_ids:
            ff = self._features[:, self._labels == unit_id]
            data.append(dict(
                x=ff[0, :].ravel(),
                y=ff[1, :].ravel(),
                mode='markers',
                name='Unit {}'.format(int(unit_id)),
                text='Unit {}'.format(int(unit_id)),
                hoverinfo='text',
                marker=dict(
                    size=self._marker_size,
                    line=dict(
                        color='white',
                        width=1
                    )
                )
            ))
        layout = dict(
            hovermode='closest'
        )
        config = dict(
            showTips=False,
            displayModeBar=True,
            displaylogo=False,
            modeBarButtonsToRemove=['hoverClosestCartesian', 'hoverCompareCartesian', 'resetScale2d']
        )
        self._plot = vd.components.PlotlyPlot(
            data=data,
            layout=layout,
            config=config,
            size=self._size
        )
        return self._plot

    def postRenderScript(self):
        js = """
        wait_for(function() {return {plotobj};}, 8, function() {
            {plotobj}.on('plotly_click', function(data){
                console.log('point clicked...', data);
            });
        });

        function wait_for(func, num_retries, cb) {
            if (func()) {
                cb();
                return;
            }
            if (num_retries<=0) return;
            setTimeout(function() {
                wait_for(func, num_retries-1, cb);
            },200);
        }
        """
        js = js.replace('{plotobj}', self._plot.javascriptPlotObject())
        # js=js.replace('{marker_size}', str(self._marker_size))
        # js=js.replace('{hover_marker_size}', str(self._hover_marker_size))
        return js

    def postRenderScript_struggling_with_hover(self):
        js = """
        setTimeout(function() {
            if ({plotobj}) {
                {plotobj}.on('plotly_hover', function(data){
                    try {
                        handle_hover(data, {hover_marker_size});
                    }
                    catch(err) {
                        console.error(err);
                        console.log('Warning: problem handling hover');
                    }
                });
                {plotobj}.on('plotly_unhover', function(data){
                    try {
                        handle_hover(data, {marker_size});
                    }
                    catch(err) {
                        console.error(err);
                        console.log('Warning: problem handling hover');
                    }
                });
                function handle_hover(data, marker_size) {
                    pt=data.points[0]
                    // color = pt.data.marker.color;
                    line = pt.data.marker.line;
                    let sizes=[];
                    for (let i=0; i<pt.data.x.length; i++)
                        sizes.push({marker_size});
                    sizes[pt.pointNumber]=marker_size;
                    Plotly.restyle({plotobj}, {marker:{size:sizes, line:line}}, [pt.curveNumber]);
                }
            }
        },100);
        """
        js = js.replace('{plotobj}', self._plot.javascriptPlotObject())
        js = js.replace('{marker_size}', str(self._marker_size))
        js = js.replace('{hover_marker_size}', str(self._hover_marker_size))
        return js


def _get_random_spike_waveforms(*, sorting, recording, unit, max_num, channels, snippet_len):
    st = sorting.get_unit_spike_train(unit_id=unit)
    num_events = len(st)
    if num_events > max_num:
        event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
    else:
        event_indices = range(num_events)

    spikes = recording.get_snippets(reference_frames=st[event_indices].astype(int), snippet_len=snippet_len, channel_ids=channels)
    if len(spikes) > 0:
        spikes = np.dstack(tuple(spikes))
    else:
        spikes = np.zeros((recording.get_num_channels(), snippet_len, 0))
    return spikes


def _get_feature_waveforms_for_spikes(*, spikes_list, num_features):
    spikes0 = spikes_list[0]
    M = spikes0.shape[0]
    T = spikes0.shape[1]
    K = len(spikes_list)
    templates = np.zeros((M, T, K))
    for k in range(K):
        templates[:, :, k] = _compute_template_from_spikes(spikes_list[k])
    if K > 1:
        print(K)
        template0 = templates[:, :, 0]
        templates2 = templates[:, :, 1:] - np.repeat(template0[:, :, np.newaxis], K - 1, axis=2)
        feature_waveforms1 = _get_feature_waveforms_helper(templates2, min(num_features, K - 1))
        if num_features <= K - 1:
            return feature_waveforms1[0:num_features, :, :]
    raise Exception('This case not yet supported')


def _get_feature_waveforms_helper(spikes, num_features):
    M = spikes.shape[0]
    T = spikes.shape[1]
    L = spikes.shape[2]
    template0 = np.mean(spikes, axis=2)
    spikes2 = spikes - np.repeat(template0[:, :, np.newaxis], L, axis=2)
    U, _, _ = svd(np.reshape(spikes2, (M * T, L)), full_matrices=False)
    return np.reshape(U[:, 0:num_features].T, (num_features, M, T))


def _compute_template_from_spikes(spikes):
    return np.median(spikes, axis=2)



import mlprocessors as mlpr
import os
import sys
from .bandpass_filter import bandpass_filter
from .whiten import whiten
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor


class MountainSort4(mlpr.Processor):
    NAME = 'MountainSort4'
    VERSION = '4.2.0'
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    CONTAINER = 'sha1://e06fee7f72f6b66d80d899ebc08e7c39e5a2458e/2019-05-06/mountainsort4.simg'
    LOCAL_MODULES = ['../../spikeforest']

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(
        'Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius = mlpr.FloatParameter(
        'Use -1 to include all channels in every neighborhood')
    freq_min = mlpr.FloatParameter(
        optional=True, default=300, description='Use 0 for no bandpass filtering')
    freq_max = mlpr.FloatParameter(
        optional=True, default=6000, description='Use 0 for no bandpass filtering')
    whiten = mlpr.BoolParameter(optional=True, default=True,
                                description='Whether to do channel whitening as part of preprocessing')
    clip_size = mlpr.IntegerParameter(
        optional=True, default=50, description='')
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=3, description='')
    detect_interval = mlpr.IntegerParameter(
        optional=True, default=10, description='Minimum number of timepoints between events detected on the same channel')
    noise_overlap_threshold = mlpr.FloatParameter(
        optional=True, default=0.15, description='Use None for no automated curation')

    def run(self):
        from .bandpass_filter import bandpass_filter
        from .whiten import whiten

        import ml_ms4alg

        print('MountainSort4......')
        recording = SFMdaRecordingExtractor(self.recording_dir)
        num_workers = os.environ.get('NUM_WORKERS', None)
        if num_workers:
            num_workers = int(num_workers)

        # Bandpass filter
        if self.freq_min or self.freq_max:
            recording = bandpass_filter(
                recording=recording, freq_min=self.freq_min, freq_max=self.freq_max)

        # Whiten
        if self.whiten:
            recording = whiten(recording=recording)

        # Sort
        sorting = ml_ms4alg.mountainsort4(
            recording=recording,
            detect_sign=self.detect_sign,
            adjacency_radius=self.adjacency_radius,
            clip_size=self.clip_size,
            detect_threshold=self.detect_threshold,
            detect_interval=self.detect_interval,
            num_workers=num_workers
        )

        # Curate
        # if self.noise_overlap_threshold is not None:
        #    sorting=ml_ms4alg.mountainsort4_curation(
        #      recording=recording,
        #      sorting=sorting,
        #      noise_overlap_threshold=self.noise_overlap_threshold
        #    )

        SFMdaSortingExtractor.write_sorting(
            sorting=sorting, save_path=self.firings_out)


class MountainSort4TestError(mlpr.Processor):
    NAME = 'MountainSort4'
    VERSION = '4.2.0'
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    CONTAINER = 'sha1://e06fee7f72f6b66d80d899ebc08e7c39e5a2458e/2019-05-06/mountainsort4.simg'

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(
        'Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius = mlpr.FloatParameter(
        'Use -1 to include all channels in every neighborhood')
    freq_min = mlpr.FloatParameter(
        optional=True, default=300, description='Use 0 for no bandpass filtering')
    freq_max = mlpr.FloatParameter(
        optional=True, default=6000, description='Use 0 for no bandpass filtering')
    whiten = mlpr.BoolParameter(optional=True, default=True,
                                description='Whether to do channel whitening as part of preprocessing')
    clip_size = mlpr.IntegerParameter(
        optional=True, default=50, description='')
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=3, description='')
    detect_interval = mlpr.IntegerParameter(
        optional=True, default=10, description='Minimum number of timepoints between events detected on the same channel')
    noise_overlap_threshold = mlpr.FloatParameter(
        optional=True, default=0.15, description='Use None for no automated curation')
    throw_error = mlpr.BoolParameter(optional=True, default=False, description='Intentionally raise an exception for purpose of testing.')

    def run(self):
        if self.throw_error:
            import time
            print('Intentionally throwing an error in 3 seconds (MountainSort4TestError)...')
            sys.stdout.flush()
            time.sleep(3)
            raise Exception('Intentional error.')
        import ml_ms4alg

        print('MountainSort4......')
        recording = SFMdaRecordingExtractor(self.recording_dir)
        num_workers = os.environ.get('NUM_WORKERS', None)
        if num_workers:
            num_workers = int(num_workers)

        # Bandpass filter
        if self.freq_min or self.freq_max:
            recording = bandpass_filter(
                recording=recording, freq_min=self.freq_min, freq_max=self.freq_max)

        # Whiten
        if self.whiten:
            recording = whiten(recording=recording)

        # Sort
        sorting = ml_ms4alg.mountainsort4(
            recording=recording,
            detect_sign=self.detect_sign,
            adjacency_radius=self.adjacency_radius,
            clip_size=self.clip_size,
            detect_threshold=self.detect_threshold,
            detect_interval=self.detect_interval,
            num_workers=num_workers
        )

        # Curate
        # if self.noise_overlap_threshold is not None:
        #    sorting=ml_ms4alg.mountainsort4_curation(
        #      recording=recording,
        #      sorting=sorting,
        #      noise_overlap_threshold=self.noise_overlap_threshold
        #    )

        SFMdaSortingExtractor.write_sorting(
            sorting=sorting, save_path=self.firings_out)



import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from spikeforest import mdaio
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import sys
import shlex
import traceback
from .install_kilosort import install_kilosort


class KiloSort(mlpr.Processor):
    """
    Kilosort wrapper for SpikeForest framework
      written by J. James Jun, May 21, 2019

    [Prerequisite]
    1. MATLAB (Tested on R2018b)
    2. CUDA Toolkit v9.1 (module load cuda/9.1.85)
    3. GCC 6.4.0 (module load gcc/6.4.0)

    [Optional: Installation instruction in SpikeForest environment]
    1. Run `git clone https://github.com/cortex-lab/KiloSort.git`
    3. In Matlab, run `CUDA/mexGPUall` to compile all CUDA codes
    4. Add `KILOSORT_PATH_DEV=...` in your .bashrc file.
    """

    NAME = 'KiloSort'
    VERSION = '0.2.3'  # wrapper VERSION
    ADDITIONAL_FILES = ['*.m']
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS']
    CONTAINER = None
    LOCAL_MODULES = ['../../spikeforest']

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(optional=True, default=-1,
                                        description='Use -1 or 1, depending on the sign of the spikes in the recording')
    adjacency_radius = mlpr.FloatParameter(optional=True, default=-1,
                                           description='Currently unused')
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=3, description='')
    # prm_template_name=mlpr.StringParameter(optional=False,description='TODO')
    freq_min = mlpr.FloatParameter(
        optional=True, default=300, description='Use 0 for no bandpass filtering')
    freq_max = mlpr.FloatParameter(
        optional=True, default=6000, description='Use 0 for no bandpass filtering')
    merge_thresh = mlpr.FloatParameter(
        optional=True, default=0.98, description='TODO')
    pc_per_chan = mlpr.IntegerParameter(
        optional=True, default=3, description='TODO')

    @staticmethod
    def install():
        print('Auto-installing kilosort.')
        return install_kilosort(
            repo='https://github.com/cortex-lab/KiloSort.git',
            commit='3f33771f8fdf8c3846a7f8a75cc8c318b44ed48c'
        )

    def run(self):
        keep_temp_files = False
        code = ''.join(random.choice(string.ascii_uppercase)
                       for x in range(10))
        tmpdir = os.environ.get('TEMPDIR', '/tmp') + '/kilosort-tmp-' + code

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)
            sorting = kilosort_helper(
                recording=recording,
                tmpdir=tmpdir,
                detect_sign=self.detect_sign,
                adjacency_radius=self.adjacency_radius,
                detect_threshold=self.detect_threshold,
                merge_thresh=self.merge_thresh,
                freq_min=self.freq_min,
                freq_max=self.freq_max,
                pc_per_chan=self.pc_per_chan
            )
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not keep_temp_files:
                    shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            shutil.rmtree(tmpdir)


def kilosort_helper(*,
                    recording,  # Recording object
                    tmpdir,  # Temporary working directory
                    detect_sign=-1,  # Polarity of the spikes, -1, 0, or 1
                    adjacency_radius=-1,  # Channel neighborhood adjacency radius corresponding to geom file
                    detect_threshold=5,  # Threshold for detection
                    merge_thresh=.98,  # Cluster merging threhold 0..1
                    freq_min=300,  # Lower frequency limit for band-pass filter
                    freq_max=6000,  # Upper frequency limit for band-pass filter
                    pc_per_chan=3,  # Number of pc per channel
                    ):
    kilosort_path = os.environ.get('KILOSORT_PATH_DEV', None)
    if kilosort_path:
        print('Using kilosort from KILOSORT_PATH_DEV directory: {}'.format(kilosort_path))
    else:
        try:
            kilosort_path = KiloSort.install()
        except:
            traceback.print_exc()
            raise Exception('Problem installing kilosort. You can set the KILOSORT_PATH_DEV to force to use a particular path.')
    print('Using kilosort from: {}'.format(kilosort_path))

    source_dir = os.path.dirname(os.path.realpath(__file__))

    dataset_dir = tmpdir + '/kilosort_dataset'
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    SFMdaRecordingExtractor.write_recording(
        recording=recording, save_path=dataset_dir, _preserve_dtype=True)

    samplerate = recording.get_sampling_frequency()

    print('Reading timeseries header...')
    HH = mdaio.readmda_header(dataset_dir + '/raw.mda')
    num_channels = HH.dims[0]
    num_timepoints = HH.dims[1]
    duration_minutes = num_timepoints / samplerate / 60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(
        num_channels, num_timepoints, duration_minutes))

    print('Creating argfile.txt file...')
    txt = ''
    txt += 'samplerate={}\n'.format(samplerate)
    txt += 'detect_sign={}\n'.format(detect_sign)
    txt += 'adjacency_radius={}\n'.format(adjacency_radius)
    txt += 'detect_threshold={}\n'.format(detect_threshold)
    txt += 'merge_thresh={}\n'.format(merge_thresh)
    txt += 'freq_min={}\n'.format(freq_min)
    txt += 'freq_max={}\n'.format(freq_max)
    txt += 'pc_per_chan={}\n'.format(pc_per_chan)
    _write_text_file(dataset_dir + '/argfile.txt', txt)

    print('Running kilosort in {tmpdir}...'.format(tmpdir=tmpdir))
    cmd = '''
        addpath('{source_dir}');
        addpath('{source_dir}/mdaio')
        try
            p_kilosort('{ksort}', '{tmpdir}', '{raw}', '{geom}', '{firings}', '{arg}');
        catch
            quit(1);
        end
        quit(0);
        '''
    cmd = cmd.format(source_dir=source_dir, ksort=kilosort_path,
                     tmpdir=tmpdir, raw=dataset_dir + '/raw.mda', geom=dataset_dir + '/geom.csv',
                     firings=tmpdir + '/firings.mda', arg=dataset_dir + '/argfile.txt')
    matlab_cmd = mlpr.ShellScript(cmd, script_path=tmpdir + '/run_kilosort.m', keep_temp_files=True)
    matlab_cmd.write()
    shell_cmd = '''
        #!/bin/bash
        cd {tmpdir}
        echo '=====================' `date` '====================='
        matlab -nosplash -nodisplay -r run_kilosort
    '''.format(tmpdir=tmpdir)
    shell_cmd = mlpr.ShellScript(shell_cmd, script_path=tmpdir + '/run_kilosort.sh', keep_temp_files=True)
    shell_cmd.write(tmpdir + '/run_kilosort.sh')
    shell_cmd.start()
    retcode = shell_cmd.wait()

    if retcode != 0:
        raise Exception('kilosort returned a non-zero exit code')

    # parse output
    result_fname = tmpdir + '/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: ' + result_fname)

    firings = mdaio.readmda(result_fname)
    sorting = se.NumpySortingExtractor()
    sorting.set_times_labels(firings[1, :], firings[2, :])
    return sorting


def _read_text_file(fname):
    with open(fname) as f:
        return f.read()


def _write_text_file(fname, str):
    with open(fname, 'w') as f:
        f.write(str)



from .spikeforest_views.currentstateview import CurrentStateView
from .spikefront_views.mainresulttableview import MainResultTableView

import vdomr as vd
from mountaintools import client as mt
import json


def get_spikefront_view_launchers(context):
    launchers = []
    groups = []
    ret = dict(
        groups=groups,
        launchers=launchers
    )

    # General
    groups.append(dict(name='general', label=''))

    launchers.append(dict(
        group='general', name='current-state', label='Current state',
        view_class=CurrentStateView,
        context=context, opts=dict(),
        enabled=True
    ))

    launchers.append(dict(
        group='general', name='main-result-table', label='Main result table',
        view_class=MainResultTableView,
        context=context, opts=dict(),
        always_open_new=False,
        enabled=True
    ))

    return ret



import numpy as np
import json
import mlprocessors as mlpr
from mountaintools import client as mt
import spikeextractors as si
from .bandpass_filter import bandpass_filter
from .sfmdaextractors import SFMdaRecordingExtractor
from .sfmdaextractors import SFMdaSortingExtractor

#_CONTAINER = 'sha1://5627c39b9bd729fc011cbfce6e8a7c37f8bcbc6b/spikeforest_basic.simg'
_CONTAINER = 'sha1://0944f052e22de0f186bb6c5cb2814a71f118f2d1/spikeforest_basic.simg' #MAY26JJJ

def write_json_file(fname, obj):
    with open(fname, 'w') as f:
        json.dump(obj, f)


def get_random_spike_waveforms(*, recording, sorting, unit, snippet_len, max_num, channels=None):
    st = sorting.get_unit_spike_train(unit_id=unit)
    num_events = len(st)
    if num_events > max_num:
        event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
    else:
        event_indices = range(num_events)

    spikes = recording.get_snippets(reference_frames=st[event_indices].astype(int), snippet_len=snippet_len, channel_ids=channels)
    if len(spikes) > 0:
        spikes = np.dstack(tuple(spikes))
    else:
        spikes = np.zeros((recording.get_num_channels(), snippet_len, 0))
    return spikes


def compute_unit_templates(*, recording, sorting, unit_ids, snippet_len=50, max_num=100, channels=None):
    ret = []
    for unit in unit_ids:
        # print('Unit {} of {}'.format(unit,len(unit_ids)))
        waveforms = get_random_spike_waveforms(recording=recording, sorting=sorting, unit=unit, snippet_len=snippet_len, max_num=max_num, channels=None)
        template = np.median(waveforms, axis=2)
        ret.append(template)
    return ret


def compute_template_snr(template, channel_noise_levels):
    channel_snrs = []
    for ch in range(template.shape[0]):
        # channel_snrs.append((np.max(template[ch,:])-np.min(template[ch,:]))/channel_noise_levels[ch])
        channel_snrs.append((np.max(np.abs(template[ch, :]))) / channel_noise_levels[ch])
    return float(np.max(channel_snrs))


def compute_channel_noise_levels(recording):
    channel_ids = recording.get_channel_ids()
    # M=len(channel_ids)
    samplerate = int(recording.get_sampling_frequency())
    X = recording.get_traces(start_frame=samplerate * 1, end_frame=samplerate * 2)
    ret = []
    for ii in range(len(channel_ids)):
        # noise_level=np.std(X[ii,:])
        noise_level = np.median(np.abs(X[ii, :])) / 0.6745  # median absolute deviation (MAD)
        ret.append(noise_level)
    return ret


class ComputeUnitsInfo(mlpr.Processor):
    NAME = 'ComputeUnitsInfo'
    VERSION = '0.1.8'
    CONTAINER = _CONTAINER
    recording_dir = mlpr.Input(directory=True, description='Recording directory')
    channel_ids = mlpr.IntegerListParameter(description='List of channels to use.', optional=True, default=[])
    unit_ids = mlpr.IntegerListParameter(description='List of units to use.', optional=True, default=[])
    firings = mlpr.Input(description='Firings file')
    json_out = mlpr.Output('The info as a .json file')

    def run(self):
        R0 = SFMdaRecordingExtractor(dataset_directory=self.recording_dir, download=True)
        sorting = SFMdaSortingExtractor(firings_file=self.firings)
        ret = compute_units_info(recording=R0, sorting=sorting, channel_ids=self.channel_ids, unit_ids=self.unit_ids)
        write_json_file(self.json_out, ret)


def compute_units_info(*, recording, sorting, channel_ids=[], unit_ids=[]):
    if (channel_ids) and (len(channel_ids) > 0):
        recording = si.SubRecordingExtractor(parent_recording=recording, channel_ids=channel_ids)

    # load into memory
    print('Loading recording into RAM...')
    recording = si.NumpyRecordingExtractor(timeseries=recording.get_traces(), samplerate=recording.get_sampling_frequency())

    # do filtering
    print('Filtering...')
    recording = bandpass_filter(recording=recording, freq_min=300, freq_max=6000)
    recording = si.NumpyRecordingExtractor(timeseries=recording.get_traces(), samplerate=recording.get_sampling_frequency())

    if (not unit_ids) or (len(unit_ids) == 0):
        unit_ids = sorting.get_unit_ids()

    print('Computing channel noise levels...')
    channel_noise_levels = compute_channel_noise_levels(recording=recording)

    # No longer use subset to compute the templates
    print('Computing unit templates...')
    templates = compute_unit_templates(recording=recording, sorting=sorting, unit_ids=unit_ids, max_num=100)

    print(recording.get_channel_ids())

    ret = []
    for i, unit_id in enumerate(unit_ids):
        print('Unit {} of {} (id={})'.format(i + 1, len(unit_ids), unit_id))
        template = templates[i]
        max_p2p_amps_on_channels = np.max(template, axis=1) - np.min(template, axis=1)
        peak_channel_index = np.argmax(max_p2p_amps_on_channels)
        peak_channel = recording.get_channel_ids()[peak_channel_index]
        peak_signal = np.max(np.abs(template[peak_channel_index, :]))
        info0 = dict()
        info0['unit_id'] = int(unit_id)
        info0['snr'] = peak_signal / channel_noise_levels[peak_channel_index]
        info0['peak_channel'] = int(recording.get_channel_ids()[peak_channel])
        train = sorting.get_unit_spike_train(unit_id=unit_id)
        info0['num_events'] = int(len(train))
        info0['firing_rate'] = float(len(train) / (recording.get_num_frames() / recording.get_sampling_frequency()))
        ret.append(info0)
    return ret


# return format can be 'json' or 'filename'
def compute_units_info_b(*, recording_dir, firings, channel_ids=[], unit_ids=[], return_format='json'):
    out = ComputeUnitsInfo.execute(
        recording_dir=recording_dir,
        firings=firings,
        unit_ids=unit_ids,
        channel_ids=channel_ids,
        json_out={'ext': '.json'},
        _container='default'
    ).outputs
    fname = out['json_out']
    if return_format == 'filename':
        return fname
    else:
        fname = mt.realizeFile(path=fname)
        with open(fname) as f:
            return json.load(f)


def select_units_on_channels(recording_dir, firings, channels):
    info = compute_units_info_b(recording_dir=recording_dir, firings=firings)
    units = []
    for info0 in info:
        if info0['peak_channel'] in channels:
            units.append(info0['unit_id'])
    return units



import vdomr as vd


class ForestViewControlPanel(vd.Component):
    def __init__(self, context):
        vd.Component.__init__(self)
        self._context = context

        self._launch_view_handlers = []
        self._context.onAnyStateChanged(self.refresh)

    def onLaunchView(self, handler):
        self._launch_view_handlers.append(handler)

    def render(self):
        view_launchers = self._context.viewLaunchers()
        groups = view_launchers['groups']
        elements = []
        for group in groups:
            view_launcher_buttons = []
            view_launcher_components = []
            group_elements = []
            for VL in view_launchers['launchers']:
                if VL['group'] == group['name']:
                    attrs = dict()
                    style0 = {'width': '130px', 'height': '28px', 'margin': '5px', 'font-size': '10px'}
                    if not VL['enabled']:
                        attrs['disabled'] = 'disabled'
                        style0['color'] = 'lightgray'
                    if 'view_class' in VL:
                        button0 = vd.components.Button(
                            label=VL['label'],
                            onclick=lambda VL=VL: self._trigger_launch_view(VL),
                            style=style0,
                            **attrs
                        )
                        view_launcher_buttons.append(button0)
                    elif 'component_class' in VL:
                        component = VL['component_class'](context=VL['context'], opts=VL['opts'])
                        view_launcher_components.append(component)
                    else:
                        print(VL)
                        raise Exception('Problem with view launcher.')
            table = _make_button_table(view_launcher_buttons, num_columns=2)
            if group['label'] is not '':
                group_elements.append(vd.h4(group['label']))
            if 'sublabel' in group:
                group_elements.append(vd.span(group['sublabel'], style={'overflow-wrap': 'break-word', 'font-size': '11px'}))
            group_elements.append(table)
            for component in view_launcher_components:
                group_elements.append(component)
            elements.append(vd.div(group_elements, style={'border': 'solid 2px lightgray', 'padding': '5px', 'margin': '5px'}))

        return vd.div(
            *elements
        )

    def _trigger_launch_view(self, VL):
        for handler in self._launch_view_handlers:
            handler(VL)


def _make_button_table(buttons, num_columns):
    rows = []
    i = 0
    while i < len(buttons):
        row_buttons = buttons[i:i + num_columns]
        while len(row_buttons) < num_columns:
            row_buttons.append(vd.span())
        rows.append(vd.tr([
            vd.td(button)
            for button in row_buttons
        ]))
        i = i + num_columns
    return vd.div(vd.table(*rows), style={'background-color': 'lightgray'})



import vdomr as vd
import time
import sys
import mtlogging
import numpy as np
from copy import deepcopy
from spikeforest import EfficientAccessRecordingExtractor
import spikeforestwidgets as SFW


class FeatureSpaceView(vd.Component):
    def __init__(self, *, context, opts=None, prepare_result=None):
        vd.Component.__init__(self)
        self._sorting_context = context
        self._recording_context = context.recordingContext()
        self._size = (100, 100)
        self._rx = EfficientAccessRecordingExtractor(path=prepare_result['earx_path'])
        unit_ids = self._sorting_context.selectedUnitIds()
        self._widget = SFW.FeatureSpaceWidgetPlotly(recording=self._rx, sorting=self._sorting_context.sortingExtractor(), unit_ids=unit_ids)
        self._widget.setSize(self._size)
        self.refresh()

    @staticmethod
    def prepareView(context, opts):
        sorting_context = context
        recording_context = context.recordingContext()
        sorting_context.initialize()
        earx = EfficientAccessRecordingExtractor(recording=recording_context.recordingExtractor())
        return dict(
            earx_path=earx.path()
        )

    def setSize(self, size):
        if self._size != size:
            self._size = size
        if self._widget:
            self._widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Feature space'

    def render(self):
        return self._widget



import mlprocessors as mlpr
import os
import time
import numpy as np
from os.path import join
from subprocess import Popen, PIPE
import shlex
import random
import string
import shutil
from spikeforest import mdaio
import spikeextractors as se
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
from mountaintools import client as mt
import json
import traceback
from .install_jrclust import install_jrclust


class JRClust(mlpr.Processor):
    """
    JRClust v4 wrapper
      written by J. James Jun, May 8, 2019
      modified from `spikeforest/spikesorters/ironclust/ironclust.py`

    [Installation instruction in SpikeForest environment]
    1. Run `git clone https://github.com/JaneliaSciComp/JRCLUST`
    2. Activate conda environment for SpikeForest
    3. Create `JRCLUST_PATH` and `MDAIO_PATH`

    See: James Jun, et al. Real-time spike sorting platform for
    high-density extracellular probes with ground-truth validation and drift correction
    https://github.com/JaneliaSciComp/JRCLUST
    """

    NAME = 'JRClust'
    VERSION = '0.1.1'
    ENVIRONMENT_VARIABLES = [
        'NUM_WORKERS', 'MKL_NUM_THREADS', 'NUMEXPR_NUM_THREADS', 'OMP_NUM_THREADS', 'TEMPDIR']
    ADDITIONAL_FILES = ['*.m', '*.prm']
    CONTAINER = None

    recording_dir = mlpr.Input('Directory of recording', directory=True)
    channels = mlpr.IntegerListParameter(
        description='List of channels to use.', optional=True, default=[])
    firings_out = mlpr.Output('Output firings file')

    detect_sign = mlpr.IntegerParameter(
        optional=True, default=-1, description='Use -1, 0, or 1, depending on the sign of the spikes in the recording')
    adjacency_radius = mlpr.FloatParameter(
        optional=True, default=50, description='')
    detect_threshold = mlpr.FloatParameter(
        optional=True, default=4.5, description='detection threshold')
    freq_min = mlpr.FloatParameter(
        optional=True, default=300, description='Use 0 for no bandpass filtering')
    freq_max = mlpr.FloatParameter(
        optional=True, default=3000, description='Use 0 for no bandpass filtering')
    merge_thresh = mlpr.FloatParameter(
        optional=True, default=0.98, description='Threshold for automated merging')
    pc_per_chan = mlpr.IntegerParameter(
        optional=True, default=1, description='Number of principal components per channel')

    # added in version 0.2.4
    filter_type = mlpr.StringParameter(
        optional=True, default='bandpass', description='{none, bandpass, wiener, fftdiff, ndiff}')
    nDiffOrder = mlpr.FloatParameter(optional=True, default=2, description='')
    common_ref_type = mlpr.StringParameter(
        optional=True, default='none', description='{none, mean, median}')
    min_count = mlpr.IntegerParameter(
        optional=True, default=30, description='Minimum cluster size')
    fGpu = mlpr.BoolParameter(
        optional=True, default=False, description='Use GPU if available')
    fParfor = mlpr.BoolParameter(
        optional=True, default=True, description='Use parfor if available')
    feature_type = mlpr.StringParameter(
        optional=True, default='gpca', description='{gpca, pca, vpp, vmin, vminmax, cov, energy, xcov}')

    def run(self):
        tmpdir = _get_tmpdir('jrclust')

        try:
            recording = SFMdaRecordingExtractor(self.recording_dir)
            params = read_dataset_params(self.recording_dir)
            if len(self.channels) > 0:
                recording = se.SubRecordingExtractor(
                    parent_recording=recording, channel_ids=self.channels)
            if not os.path.exists(tmpdir):
                os.mkdir(tmpdir)

            all_params = dict()
            for param0 in self.PARAMETERS:
                all_params[param0.name] = getattr(self, param0.name)
            sorting = jrclust_helper(
                recording=recording,
                tmpdir=tmpdir,
                params=params,
                **all_params,
            )
            SFMdaSortingExtractor.write_sorting(
                sorting=sorting, save_path=self.firings_out)
        except:
            if os.path.exists(tmpdir):
                if not getattr(self, '_keep_temp_files', False):
                    shutil.rmtree(tmpdir)
            raise
        if not getattr(self, '_keep_temp_files', False):
            shutil.rmtree(tmpdir)


def jrclust_helper(
        *,
        recording,  # Recording object
        tmpdir,  # Temporary working directory
        params=dict(),
        **kwargs):

    jrclust_path = os.environ.get('JRCLUST_PATH_DEV', None)
    if jrclust_path:
        print('Using jrclust from JRCLUST_PATH_DEV directory: {}'.format(jrclust_path))
    else:
        try:
            print('Auto-installing jrclust.')
            jrclust_path = install_jrclust(
                repo='https://github.com/JaneliaSciComp/JRCLUST.git',
                commit='68ffb3ef064f97aca7043b7faac49c34a58997d9'
            )
        except:
            traceback.print_exc()
            raise Exception('Problem installing jrclust. You can set the JRCLUST_PATH_DEV to force to use a particular path.')
    print('Using jrclust from: {}'.format(jrclust_path))

    dataset_dir = os.path.join(tmpdir, 'jrclust_dataset')
    # Generate three files in the dataset directory: raw.mda, geom.csv, params.json
    SFMdaRecordingExtractor.write_recording(
        recording=recording, save_path=dataset_dir, params=params, _preserve_dtype=True)

    samplerate = recording.get_sampling_frequency()

    print('Reading timeseries header...')
    raw_mda = os.path.join(dataset_dir, 'raw.mda')
    HH = mdaio.readmda_header(raw_mda)
    num_channels = HH.dims[0]
    num_timepoints = HH.dims[1]
    duration_minutes = num_timepoints / samplerate / 60
    print('Num. channels = {}, Num. timepoints = {}, duration = {} minutes'.format(
        num_channels, num_timepoints, duration_minutes))

    print('Creating argfile.txt...')
    txt = ''
    for key0, val0 in kwargs.items():
        txt += '{}={}\n'.format(key0, val0)
    if 'scale_factor' in params:
        txt += 'bitScaling={}\n'.format(params["scale_factor"])
    txt += 'sampleRate={}\n'.format(samplerate)
    _write_text_file(dataset_dir + '/argfile.txt', txt)

    # new method
    source_path = os.path.dirname(os.path.realpath(__file__))
    print('Running jrclust in {tmpdir}...'.format(tmpdir=tmpdir))
    cmd = '''
        addpath('{jrclust_path}', '{source_path}', '{source_path}/mdaio');
        try
            p_jrclust('{tmpdir}', '{dataset_dir}/raw.mda', '{dataset_dir}/geom.csv', '{tmpdir}/firings.mda', '{dataset_dir}/argfile.txt');
        catch
            fprintf('----------------------------------------');
            fprintf(lasterr());
            quit(1);
        end
        quit(0);
    '''
    cmd = cmd.format(jrclust_path=jrclust_path, tmpdir=tmpdir, dataset_dir=dataset_dir, source_path=source_path)

    matlab_cmd = mlpr.ShellScript(cmd, script_path=tmpdir + '/run_jrclust.m', keep_temp_files=True)
    matlab_cmd.write()

    shell_cmd = '''
        #!/bin/bash
        cd {tmpdir}
        matlab -nosplash -nodisplay -r run_jrclust
    '''.format(tmpdir=tmpdir)
    shell_cmd = mlpr.ShellScript(shell_cmd, script_path=tmpdir + '/run_jrclust.sh', keep_temp_files=True)
    shell_cmd.write(tmpdir + '/run_jrclust.sh')
    shell_cmd.start()

    retcode = shell_cmd.wait()

    if retcode != 0:
        raise Exception('jrclust returned a non-zero exit code')

    # parse output
    result_fname = tmpdir + '/firings.mda'
    if not os.path.exists(result_fname):
        raise Exception('Result file does not exist: ' + result_fname)

    firings = mdaio.readmda(result_fname)
    sorting = se.NumpySortingExtractor()
    sorting.set_times_labels(firings[1, :], firings[2, :])
    return sorting


def _read_text_file(fname):
    with open(fname) as f:
        return f.read()


def _write_text_file(fname, str):
    with open(fname, 'w') as f:
        f.write(str)


def _run_command_and_print_output(command):
    with Popen(shlex.split(command), stdout=PIPE, stderr=PIPE) as process:
        while True:
            output_stdout = process.stdout.readline()
            output_stderr = process.stderr.readline()
            if (not output_stdout) and (not output_stderr) and (process.poll() is not None):
                break
            if output_stdout:
                print(output_stdout.decode())
            if output_stderr:
                print(output_stderr.decode())
        rc = process.poll()
        return rc


def read_dataset_params(dsdir):
    # ca = _load_required_modules()
    fname1 = dsdir + '/params.json'
    fname2 = mt.realizeFile(path=fname1)
    if not fname2:
        raise Exception('Unable to find file: ' + fname1)
    if not os.path.exists(fname2):
        raise Exception('Dataset parameter file does not exist: ' + fname2)
    with open(fname2) as f:
        return json.load(f)


# To be shared across sorters (2019.05.05)
def _get_tmpdir(sorter_name):
    code = ''.join(random.choice(string.ascii_uppercase) for x in range(10))
    tmpdir0 = os.environ.get('TEMPDIR', '/tmp')
    tmpdir = os.path.join(tmpdir0, '{}-tmp-{}'.format(sorter_name, code))
    # reset the output folder
    if os.path.exists(tmpdir):
        shutil.rmtree(str(tmpdir))
    else:
        os.makedirs(tmpdir)
    return tmpdir



import uuid
from spikeforest import mdaio
import io
import base64
import vdomr as vd
import os
from matplotlib import pyplot as plt
from sklearn.decomposition import PCA
import numpy as np


class FeatureSpaceWidgetPlotly(vd.Component):
    def __init__(self, *, recording, sorting, channels=None, unit_ids=None, width=14, height=7, snippet_len=100,
                 title='', max_num_spikes_per_unit=1000):
        vd.Component.__init__(self)
        self._recording = recording
        self._sorting = sorting
        self._channels = channels
        self._unit_ids = unit_ids
        self._width = width
        self._height = height
        self._figure = None
        self._features = []
        self._snippet_len = snippet_len
        self._title = title
        self._max_num_spikes_per_unit = max_num_spikes_per_unit
        self._size = (800, 500)
        self._do_compute()

    def setSize(self, size):
        if self._size == size:
            return
        self._size = size
        self.refresh()

    def render(self):
        data = []
        for ii, ff in enumerate(self._features):
            data.append(dict(
                x=ff[1, :].ravel(),
                y=ff[2, :].ravel(),
                mode='markers',
                name='Unit {}'.format(self._unit_ids[ii])
            ))
        plot = vd.components.PlotlyPlot(
            data=data,
            layout=dict(margin=dict(t=5)),
            config=dict(),
            size=self._size
        )
        return plot

    def plot(self):
        self._do_plot()

    def figure(self):
        return self._figure

    def _do_compute(self):
        units = self._unit_ids
        channels = self._channels
        if units is None:
            units = self._sorting.get_unit_ids()
        channel_ids = self._recording.get_channel_ids()
        M = len(channel_ids)
        channel_locations = np.zeros((M, 2))
        for ii, ch in enumerate(channel_ids):
            loc = self._recording.get_channel_property(ch, 'location')
            channel_locations[ii, :] = loc[-2:]
        if channels is None:
            channels = channel_ids
        self._spike_waveforms_list = []
        for unit in units:
            st = self._sorting.get_unit_spike_train(unit_id=unit)
            if st is not None:
                spikes, spiketimes = self._get_random_spike_waveforms(unit=unit, max_num=self._max_num_spikes_per_unit,
                                                                      channels=channels)
                item = dict(
                    representative_spiketimes=spiketimes,
                    representative_waveforms=spikes,
                    title='Unit {}'.format(int(unit))
                )
                self._spike_waveforms_list.append(item)
            else:
                print(unit, ' spike train is None')
        self._opts = {'channel': -1,
                      'feature_extraction': 'PCA',
                      'features_to_show': (2, 1)}
        self._get_spikes_in_feature_space_multi(self._spike_waveforms_list, self._opts)

    def _do_plot(self):
        self._do_compute()
        with plt.rc_context({'axes.edgecolor': 'gray'}):
            self._plot_spikes_in_feature_space_multi(self._spike_waveforms_list, self._opts)

    def _get_spikes_in_feature_space_multi(self, list, opts):
        channel = opts['channel']
        all_waveforms = None
        for spikes_dict in list:
            waveforms = spikes_dict['representative_waveforms']
            if channel == -1:
                s = np.shape(waveforms)
                spikes_dict['representative_waveforms'] = waveforms = waveforms.reshape(s[0] * s[1], s[2])
            else:
                spikes_dict['representative_waveforms'] = waveforms = waveforms[channel, :, :]
            if all_waveforms is None:
                all_waveforms = waveforms
            else:
                all_waveforms = np.concatenate([all_waveforms, waveforms], axis=1)
        feature_space = self._compute_features(all_waveforms, opts)
        for d in list:
            self._get_spikes_in_feature_space(d, feature_space, opts)

    def _get_spikes_in_feature_space(self, spikes_dict, feature_space, opts):
        waveforms = spikes_dict['representative_waveforms']
        spiketimes = spikes_dict['representative_spiketimes']
        if not (np.shape(waveforms)[1] > 0):
            raise Exception('{} does not exist or has no spikes.'.format(spikes_dict['title']))
        features = feature_space.transform(waveforms.T).T
        spiketimes = (np.array(spiketimes) - np.max(spiketimes) / 2) / 2
        spiketimes.shape = (1, len(spiketimes))
        features = np.concatenate([spiketimes, features], axis=0)
        self._features.append(features)
        # TODO: Optionally exclude outliers

    def _plot_spikes_in_feature_space_multi(self, list, opts):
        features_to_show = opts['features_to_show']
        _ = plt.figure(figsize=(10, 10))
        for i, fet in enumerate(self._features):
            plt.scatter(fet[:, features_to_show[0]], fet[:, features_to_show[1]],
                        label=list[i]['title'])
        plt.legend()
        plt.xlabel('{} - {}'.format(opts['feature_extraction'], opts['features_to_show'][0]))
        plt.ylabel('{} - {}'.format(opts['feature_extraction'], opts['features_to_show'][1]))

    def _compute_features(self, spike_waveforms, opts):
        if opts['feature_extraction'] == 'PCA':
            features_obj = PCA()
            features_obj.fit(spike_waveforms.T)
        else:
            raise NotImplementedError('Override _compute_features if feature extraction other than PCA is required')
        return features_obj

    def _get_random_spike_waveforms(self, *, unit, max_num, channels):
        st = self._sorting.get_unit_spike_train(unit_id=unit)
        num_events = len(st)
        if num_events > max_num:
            event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
        else:
            event_indices = range(num_events)

        spikes = self._recording.get_snippets(reference_frames=st[event_indices].astype(int), snippet_len=self._snippet_len, channel_ids=channels)
        if len(spikes) > 0:
            spikes = np.dstack(tuple(spikes))
        else:
            spikes = np.zeros((self._recording.get_num_channels(), self._snippet_len, 0))
        return spikes, st[event_indices]



import numpy as np


def get_random_spike_waveforms(*, recording, sorting, unit, snippet_len, max_num, channels=None):
    st = sorting.get_unit_spike_train(unit_id=unit)
    num_events = len(st)
    if num_events > max_num:
        event_indices = np.random.choice(range(num_events), size=max_num, replace=False)
    else:
        event_indices = range(num_events)

    spikes = recording.get_snippets(reference_frames=st[event_indices].astype(int), snippet_len=snippet_len, channel_ids=channels)
    spikes = np.dstack(tuple(spikes))
    return spikes


def compute_unit_templates(*, recording, sorting, unit_ids, snippet_len=50):
    ret = []
    for unit in unit_ids:
        waveforms = get_random_spike_waveforms(recording=recording, sorting=sorting, unit=unit, snippet_len=snippet_len,
                                               max_num=200)
        template = np.mean(waveforms, axis=2)
        ret.append(template)
    return ret


def compute_template_snr(template, channel_noise_levels):
    channel_snrs = []
    for ch in range(template.shape[0]):
        channel_snrs.append((np.max(template[ch, :]) - np.min(template[ch, :])) / channel_noise_levels[ch])
    return np.max(channel_snrs)


def compute_channel_noise_levels(recording):
    channel_ids = recording.get_channel_ids()
    # M = len(channel_ids)
    X = recording.get_traces(start_frame=0, end_frame=np.minimum(1000, recording.get_num_frames()))
    ret = []
    for ii in len(channel_ids):
        noise_level = np.std(X[ii, :])
        ret.append(noise_level)
    return ret


def compute_unit_snrs(*, recording, sorting, unit_ids=None):
    if unit_ids is None:
        unit_ids = sorting.get_unit_ids()
    channel_noise_levels = compute_channel_noise_levels(recording=recording)
    templates = compute_unit_templates(recording=recording, sorting=sorting, unit_ids=unit_ids)
    ret = []
    for template in templates:
        snr = compute_template_snr(template, channel_noise_levels)
        ret.append(snr)
    return ret



from matplotlib import pyplot as plt
import numpy as np


class ElectrodeGeometryWidget:
    def __init__(self, *, recording):
        self._recording = recording

    def plot(self, width=1.5, height=1.5):
        self._do_plot(width=width, height=height)

    def _get_geom(self):
        RX = self._recording
        return np.stack([RX.get_channel_property(channel_id=ch, property_name='location') for ch in RX.get_channel_ids()])

    def _do_plot(self, width, height):
        # R = self._recording
        geom = self._get_geom()

        fig = plt.figure(figsize=(width, height))
        ax = fig.add_axes([0, 0, 1, 1])
        ax.axis('off')

        x = geom[:, 0]
        y = geom[:, 1]
        xmin = np.min(x); xmax = np.max(x)
        ymin = np.min(y); ymax = np.max(y)

        marker_size = width * fig.dpi / 6
        margin = np.maximum(xmax - xmin, ymax - ymin) * 0.2

        plt.scatter(x, y, marker='o', s=int(marker_size**2))
        plt.axis('equal')
        plt.xticks([])
        plt.yticks([])
        plt.xlim(xmin - margin, xmax + margin)
        plt.ylim(ymin - margin, ymax + margin)
        # plt.show()



import vdomr as vd
import time
import sys
import mtlogging
import numpy as np
import json
from .tablewidget import TableWidget


class RecordingTableView(vd.Component):
    def __init__(self, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._recording_table_widget = RecordingTableWidget(context=context)

    def setSize(self, size):
        self._size = size
        if self._recording_table_widget:
            self._recording_table_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Recording table'

    def render(self):
        return vd.div(
            self._recording_table_widget
        )

# class RecordingTableWidget(vd.Component):
#     def __init__(self, *, context):
#         vd.Component.__init__(self)
#         self._context = context
#         self._size = (100,100)
#     def setSize(self, size):
#         self._size = size
#         self.refresh()
#     def render(self):
#         return vd.pre(json.dumps(self._context.recordingNames(), indent=4))


class RecordingTableWidget(vd.Component):
    def __init__(self, *, context):
        vd.Component.__init__(self)
        self._size = (100, 100)
        self._context = context
        self._update_table()
        self._context.onCurrentRecordingChanged(self._on_context_selection_changed)

    def setSize(self, size):
        self._size = size
        self._update_table()

    def _update_table(self):
        self._recording_ids = self._context.recordingIds()
        recobjs = [self._context.recordingObject(recid) for recid in self._recording_ids]
        records = [
            dict(
                study_name=obj['study'],
                recording_name=obj['name'],
                duration_sec=_get_computed_info(obj, 'duration_sec'),
                samplerate=_get_computed_info(obj, 'samplerate'),
                num_channels=_get_computed_info(obj, 'num_channels')
            )
            for obj in recobjs
        ]

        self._table_widget = TableWidget(
            columns=[
                dict(label='Study', name='study_name'),
                dict(label='Recording', name='recording_name'),
                dict(label='Duration (sec)', name='duration_sec'),
                dict(label='Sampling freq. (Hz)', name='samplerate'),
                dict(label='Num. channels', name='num_channels'),
            ],
            records=records,
            height=self._size[1]
        )
        self._table_widget.onSelectionChanged(self._on_widget_selection_changed)
        self._on_context_selection_changed()
        self.refresh()

    def _on_widget_selection_changed(self):
        current_row_index = self._table_widget.currentRowIndex()
        if current_row_index is not None:
            self._context.setCurrentRecordingId(self._recording_ids[current_row_index])
        else:
            self._context.setCurrentRecordingId(None)

    def _on_context_selection_changed(self):
        recid = self._context.currentRecordingId()
        if recid is not None:
            try:
                index0 = self._recording_ids.index(recid)
            except:
                index0 = None
        else:
            index0 = None
        self._table_widget.setCurrentRowIndex(index0)

    def render(self):
        return self._table_widget


class RecordingSelectComponent(vd.Component):
    def __init__(self, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._recording_select_widget = RecordingSelectWidget(context=context)

    def render(self):
        return self._recording_select_widget


class RecordingSelectWidget(vd.Component):
    def __init__(self, *, context):
        vd.Component.__init__(self)
        self._context = context
        self._select_box = vd.components.SelectBox(style={'width': '100%', 'font-size': '12px'})
        self._select_box.onChange(self._on_selection_changed)
        self._context.onCurrentRecordingChanged(self._on_context_selection_changed)
        self._update_options()

    def _update_options(self):
        self._recording_ids = self._context.recordingIds()
        self._select_box.setOptions(['[Select recording]'] + self._recording_ids)
        recid = self._context.currentRecordingId()
        if recid is not None:
            self._select_box.setValue(recid)

        self._on_context_selection_changed()
        self.refresh()

    def _on_selection_changed(self, value):
        index0 = self._select_box.index()
        if index0 > 0:
            recid = self._recording_ids[index0 - 1]
            self._context.setCurrentRecordingId(recid)

    def _on_context_selection_changed(self):
        recid = self._context.currentRecordingId()
        self._select_box.setValue(recid)

    def render(self):
        return self._select_box


def _get_computed_info(recobj, name):
    summary = recobj.get('summary', dict())
    computed_info = summary.get('computed_info', dict())
    return computed_info.get(name, None)



# pylint: disable=no-member

import mlprocessors as mlpr
import spikeextractors as se
import h5py
import numpy as np
import spikeextractors as se
from mountaintools import client as mt
import mtlogging


class EfficientAccessRecordingExtractor(se.RecordingExtractor):
    def __init__(self, *, path=None, recording=None, _dest_path=None):
        se.RecordingExtractor.__init__(self)
        if path is not None:
            if recording is not None:
                raise Exception('Cannot pass both path and recording to EfficientAccessRecordingExtractor.')
            self._path = path
        elif recording is not None:
            if not hasattr(recording, 'hash'):
                try:
                    if not hasattr(recording, 'hash'):
                        print('''
                        Warning: Recording does not have the hash attribute.
                        Using sampling method to compute a hash.''')
                        setattr(recording, 'hash', _hash(recording))
                except:
                    raise Exception('Recording class does not support the sampling hash.')
            path0 = CreateEfficientAccessRecordingFile.execute(
                recording=recording,
                hdf5_out=dict(ext='.hdf5', dest_path=_dest_path)
            ).outputs['hdf5_out']
            self._path = mt.realizeFile(path=path0)
        else:
            raise Exception('Missing argument: path or recording')

        with h5py.File(self._path, "r") as f:
            self._num_segments = int(np.array(f.get('num_segments'))[0])
            self._segment_size = int(np.array(f.get('segment_size'))[0])
            self._num_channels = int(np.array(f.get('num_channels'))[0])
            self._channel_ids = np.array(f.get('channel_ids')).tolist()
            self._num_timepoints = int(np.array(f.get('num_timepoints'))[0])
            self._samplerate = np.array(f.get('samplerate'))[0]
            self._recording_hash = np.array(f.get('recording_hash'))[0].decode()
            assert type(self._recording_hash) == str
            geom = np.array(f.get('geom'))
            channel_locations = [geom[m, :].ravel() for m in range(self._num_channels)]
            self.set_channel_locations(channel_ids=self._channel_ids, locations=channel_locations)

    def hash(self):
        return self._recording_hash

    def path(self):
        return self._path

    def get_channel_ids(self):
        return self._channel_ids

    def get_num_frames(self):
        return self._num_timepoints

    def get_sampling_frequency(self):
        return self._samplerate

    @mtlogging.log(name='EfficientAccessRecordingExtractor:get_traces')
    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None):
        if start_frame is None:
            start_frame = 0
        if end_frame is None:
            end_frame = self.get_num_frames()
        if channel_ids is None:
            channel_ids = self.get_channel_ids()

        t1 = int(start_frame)
        t2 = int(end_frame)
        c1 = int(t1 / self._segment_size)
        c2 = int((t2 - 1) / self._segment_size)
        ret = np.zeros((len(channel_ids), t2 - t1))
        with h5py.File(self._path, "r") as f:
            for cc in range(c1, c2 + 1):
                if cc == c1:
                    t1a = t1
                else:
                    t1a = self._segment_size * cc
                if cc == c2:
                    t2a = t2
                else:
                    t2a = self._segment_size * (cc + 1)
                for ii, m in enumerate(channel_ids):
                    str0 = 'part-{}-{}'.format(m, cc)
                    offset = self._segment_size * cc
                    ret[ii, t1a - t1:t2a - t1] = f[str0][t1a - offset:t2a - offset]
        return ret

    @staticmethod
    def write_recording(recording, save_path):
        EfficientAccessRecordingExtractor(recording=recording, _dest_path=save_path)


class CreateEfficientAccessRecordingFile(mlpr.Processor):
    NAME = 'CreateEfficientAccessRecordingFile'
    VERSION = '0.1.2'
    recording = mlpr.Input()
    segment_size = mlpr.IntegerParameter(optional=True, default=1000000)
    hdf5_out = mlpr.Output()

    def run(self):
        recording = self.recording
        segment_size = self.segment_size
        channel_ids = recording.get_channel_ids()
        samplerate = recording.get_sampling_frequency()
        M = len(channel_ids)  # number of channels
        N = recording.get_num_frames()  # Number of timepoints
        num_segments = int(np.ceil(N / segment_size))

        channel_locations = recording.get_channel_locations(channel_ids=channel_ids)
        nd = len(channel_locations[0])
        geom = np.zeros((M, nd))
        for m in range(M):
            geom[m, :] = channel_locations[m]

        with h5py.File(self.hdf5_out, "w") as f:
            f.create_dataset('segment_size', data=[segment_size])
            f.create_dataset('num_segments', data=[num_segments])
            f.create_dataset('num_channels', data=[M])
            f.create_dataset('channel_ids', data=np.array(channel_ids))
            f.create_dataset('num_timepoints', data=[N])
            f.create_dataset('samplerate', data=[samplerate])
            f.create_dataset('geom', data=geom)
            if callable(recording.hash):
                hash0 = recording.hash()
            else:
                hash0 = recording.hash
            f.create_dataset('recording_hash', data=np.array([hash0.encode()]))
            for j in range(num_segments):
                segment = np.zeros((M, segment_size), dtype=float)  # fix dtype here
                t1 = int(j * segment_size)  # first timepoint of the segment
                t2 = int(np.minimum(N, (t1 + segment_size)))  # last timepoint of segment (+1)
                s1 = int(np.maximum(0, t1))  # first timepoint
                s2 = int(np.minimum(N, t2))  # last timepoint (+1)

                # determine aa so that t1-s1+aa = 0
                # so, aa = -(t1-s1)
                aa = -(t1 - s1)
                segment[:, aa:aa + s2 - s1] = recording.get_traces(start_frame=s1, end_frame=s2)  # Read the segment

                for ii, ch in enumerate(channel_ids):
                    f.create_dataset('part-{}-{}'.format(ch, j), data=segment[ii, :].ravel())


def _hash(self):
    from mountaintools import client as mt
    obj = {
        'channels': tuple(self.get_channel_ids()),
        'frames': self.get_num_frames(),
        'data': samplehash(self)
    }
    return mt.sha1OfObject(obj)


def samplehash(self):
    rng = np.random.RandomState(37)
    n_samples = min(self.get_num_frames() // 1000, 100)
    inds = rng.randint(low=0, high=self.get_num_frames(), size=n_samples)
    h = 0
    for i in inds:
        t = self.get_traces(start_frame=i, end_frame=i + 100)
        h = hash((hash(bytes(t)), hash(h)))
    return h



#!/usr/bin/env python

import spikeforest_analysis as sa
from mountaintools import client as mt
from spikeforest import example_datasets
from spikeforest import SFMdaRecordingExtractor, SFMdaSortingExtractor
import os
import shutil
import sfdata as sf
import numpy as np
import pytest

# This file runs a spikeforest processing pipeline locally (no internet connection, except for downloading singularity containers if needed)
#    * Generate synthetic recordings (toy recordings)
#    * Summarize the recordings
#    * Sort the recordings
#    * Summarize the sortings
#    * Compare with ground truth
#    * Save results to kbucket (local) for later retrieval by GUIs, website, etc
#
#    NOTE: For now the singularity containers need to be on your system --
#          will improve this functionality in the future
#
#    Modify the settings below and then run this file using python 3


@pytest.mark.slow
@pytest.mark.analysis
def test_spikeforest_analysis(tmpdir):
    tmpdir = str(tmpdir)

    # generate toy recordings
    delete_recordings = True
    num_recordings = 2
    duration = 15
    for num in range(1, num_recordings + 1):
        dirname = tmpdir + '/toy_example{}'.format(num)
        if delete_recordings:
            if os.path.exists(dirname):
                shutil.rmtree(dirname)
        if not os.path.exists(dirname):
            rx, sx_true = example_datasets.toy_example1(
                duration=duration, num_channels=4, samplerate=30000, K=10)
            SFMdaRecordingExtractor.write_recording(
                recording=rx, save_path=dirname)
            SFMdaSortingExtractor.write_sorting(
                sorting=sx_true, save_path=dirname + '/firings_true.mda')

    # # Use this to optionally connect to a kbucket share:
    # # ca.autoConfig(collection='spikeforest',key='spikeforest2-readwrite',ask_password=True)
    # # for downloading containers if needed
    mt.configDownloadFrom(['spikeforest.spikeforest2'])

    # Specify the compute resource (see the note above)
    # compute_resource = 'local-computer'
    compute_resource = None

    # Use this to control whether we force the processing to re-run (by default it uses cached results)
    os.environ['MLPROCESSORS_FORCE_RUN'] = 'FALSE'  # FALSE or TRUE

    # This is the id of the output -- for later retrieval by GUI's, etc
    output_id = 'spikeforest_test0'

    # Grab the recordings for testing
    recordings = [
        dict(
            recording_name='toy_example{}'.format(num),
            study_name='toy_examples',
            directory=tmpdir + '/toy_example{}'.format(num)
        )
        for num in range(1, num_recordings + 1)
    ]

    studies = [
        dict(
            name='toy_examples',
            study_set='toy_examples',
            directory=os.path.abspath('.'),
            description='Toy examples.'
        )
    ]

    # Summarize the recordings
    recordings = sa.summarize_recordings(
        recordings=recordings, compute_resource=compute_resource)

    # Sorters (algs and params) are defined below
    sorters = _define_sorters()

    # We will be assembling the sorting results here
    sorting_results = []

    for sorter in sorters:
        # Sort the recordings
        sortings = sa.sort_recordings(
            sorter=sorter,
            recordings=recordings,
            compute_resource=compute_resource
        )

        # Summarize the sortings
        sortings = sa.summarize_sortings(
            sortings=sortings,
            compute_resource=compute_resource
        )

        # Compare with ground truth
        sortings = sa.compare_sortings_with_truth(
            sortings=sortings,
            compute_resource=compute_resource
        )

        # Append to results
        sorting_results = sorting_results + sortings

    # TODO: collect all the units for aggregated analysis

    aggregated_sorting_results = sa.aggregate_sorting_results(studies, recordings, sorting_results)

    # Save the output
    print('Saving the output')
    mt.saveObject(
        key=dict(
            name='spikeforest_results',
            output_id=output_id
        ),
        object=dict(
            studies=studies,
            recordings=recordings,
            sorting_results=sorting_results,
            aggregated_sorting_results=mt.saveObject(object=aggregated_sorting_results)
        )
    )

    for sr in aggregated_sorting_results['study_sorting_results']:
        study_name = sr['study']
        sorter_name = sr['sorter']
        n1 = np.array(sr['num_matches'])
        n2 = np.array(sr['num_false_positives'])
        n3 = np.array(sr['num_false_negatives'])
        accuracies = n1 / (n1 + n2 + n3)
        avg_accuracy = np.mean(accuracies)
        txt = 'STUDY: {}, SORTER: {}, AVG ACCURACY: {}'.format(study_name, sorter_name, avg_accuracy)
        print(txt)
        if avg_accuracy < 0.3:
            if sorter_name == 'Yass':
                print('Average accuracy is too low, but we are excusing Yass for now.')
            else:
                raise Exception('Average accuracy is too low for test----- ' + txt)


def _define_sorters():
    sorter_ms4_thr3 = dict(
        name='MountainSort4-thr3',
        processor_name='MountainSort4',
        params=dict(
            detect_sign=-1,
            adjacency_radius=50,
            detect_threshold=3.1
        )
    )

    sorter_sc = dict(
        name='SpykingCircus',
        processor_name='SpykingCircus',
        params=dict(
            detect_sign=-1,
            adjacency_radius=50
        )
    )

    sorter_yass = dict(
        name='Yass',
        processor_name='Yass',
        params=dict(
            detect_sign=-1,
            adjacency_radius=50
        )
    )
    return [sorter_ms4_thr3, sorter_sc, sorter_yass]
    # return [sorter_ms4_thr3]


import vdomr as vd
import json


class CurrentStateView(vd.Component):
    def __init__(self, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._context.onAnyStateChanged(self.refresh)
        self._size = (100, 100)

    def tabLabel(self):
        return 'Current state'

    def setSize(self, size):
        if self._size == size:
            return
        self._size = size
        self.refresh()

    def size(self):
        return self._size

    def render(self):
        state0 = self._context.stateObject()
        return vd.div(
            vd.pre(
                json.dumps(state0, indent=4)
            )
        )



import numpy as np
from spikeextractors import RecordingExtractor, SortingExtractor, SubRecordingExtractor, SubSortingExtractor
import csv
from pathlib import Path


def read_python(path):
    '''Parses python scripts in a dictionary

    Parameters
    ----------
    path: str
        Path to file to parse

    Returns
    -------
    metadata:
        dictionary containing parsed file

    '''
    from six import exec_
    path = Path(path).absolute()
    assert path.is_file()
    with path.open('r') as f:
        contents = f.read()
    metadata = {}
    exec_(contents, {}, metadata)
    metadata = {k.lower(): v for (k, v) in metadata.items()}
    return metadata


def loadProbeFile(recording, probe_file, channel_map=None, channel_groups=None):
    '''Loads channel information into recording extractor. If a .prb file is given,
    then 'location' and 'group' information for each channel is stored. If a .csv
    file is given, then it will only store 'location'

    Parameters
    ----------
    recording: RecordingExtractor
        The recording extractor to channel information
    probe_file: str
        Path to probe file. Either .prb or .csv
    Returns
    ---------
    subRecordingExtractor
    '''
    probe_file = Path(probe_file)
    if probe_file.suffix == '.prb':
        probe_dict = read_python(probe_file)
        if 'channel_groups' in probe_dict.keys():
            ordered_channels = np.array([], dtype=int)
            groups = sorted(probe_dict['channel_groups'].keys())
            for cgroup_id in groups:
                cgroup = probe_dict['channel_groups'][cgroup_id]
                for key_prop, prop_val in cgroup.items():
                    if key_prop == 'channels':
                        ordered_channels = np.concatenate((ordered_channels, prop_val))

            if list(ordered_channels) == recording.get_channel_ids():
                subrecording = recording
            else:
                assert np.all([chan in recording.get_channel_ids() for chan in ordered_channels]), \
                    "all channel_ids in the 'channels' section of the probe file " \
                    "must be in the original recording channel ids"
                subrecording = SubRecordingExtractor(recording, channel_ids=list(ordered_channels))
            for cgroup_id in groups:
                cgroup = probe_dict['channel_groups'][cgroup_id]
                if 'channels' not in cgroup.keys() and len(groups) > 1:
                    raise Exception("If more than one 'channel_group' is in the probe file, the 'channels' field"
                                    "for each channel group is required")
                elif 'channels' not in cgroup.keys():
                    channels_in_group = subrecording.get_num_channels()
                else:
                    channels_in_group = len(cgroup['channels'])
                for key_prop, prop_val in cgroup.items():
                    if key_prop == 'channels':
                        for i_ch, prop in enumerate(prop_val):
                            subrecording.set_channel_property(prop, 'group', int(cgroup_id))
                    elif key_prop == 'geometry' or key_prop == 'location':
                        if isinstance(prop_val, dict) and len(prop_val.keys()) == channels_in_group:
                            for (i_ch, prop) in prop_val.items():
                                subrecording.set_channel_property(i_ch, 'location', prop)
                        elif isinstance(prop_val, (list, np.ndarray)) and len(prop_val) == channels_in_group:
                            for (i_ch, prop) in zip(subrecording.get_channel_ids(), prop_val):
                                subrecording.set_channel_property(i_ch, 'location', prop)
                    else:
                        if isinstance(prop_val, dict) and len(prop_val.keys()) == channels_in_group:
                            for (i_ch, prop) in prop_val.items():
                                subrecording.set_channel_property(i_ch, key_prop, prop)
                        elif isinstance(prop_val, (list, np.ndarray)) and len(prop_val) == channels_in_group:
                            for (i_ch, prop) in zip(subrecording.get_channel_ids(), prop_val):
                                subrecording.set_channel_property(i_ch, key_prop, prop)
                # create dummy locations
                if 'geometry' not in cgroup.keys() and 'location' not in cgroup.keys():
                    for i, chan in enumerate(subrecording.get_channel_ids()):
                        subrecording.set_channel_property(chan, 'location', [i, 0])
        else:
            raise AttributeError("'.prb' file should contain the 'channel_groups' field")

    elif probe_file.suffix == '.csv':
        if channel_map is not None:
            assert np.all([chan in recording.get_channel_ids() for chan in channel_map]), \
                "all channel_ids in 'channel_map' must be in the original recording channel ids"
            subrecording = SubRecordingExtractor(recording, channel_ids=channel_map)
        else:
            subrecording = recording
        with open(probe_file) as csvfile:
            posreader = csv.reader(csvfile)
            row_count = 0
            loaded_pos = []
            for pos in (posreader):
                row_count += 1
                loaded_pos.append(pos)
            assert len(subrecording.get_channel_ids()) == row_count, "The .csv file must contain as many rows as the number of channels in the recordings"
            for i_ch, pos in zip(subrecording.get_channel_ids(), loaded_pos):
                subrecording.set_channel_property(i_ch, 'location', list(np.array(pos).astype(float)))
            if channel_groups is not None and len(channel_groups) == len(subrecording.get_channel_ids()):
                for i_ch, chg in zip(subrecording.get_channel_ids(), channel_groups):
                    subrecording.set_channel_property(i_ch, 'group', chg)
    else:
        raise NotImplementedError("Only .csv and .prb probe files can be loaded.")

    return subrecording


def saveProbeFile(recording, probe_file, format=None, radius=100, dimensions=None):
    '''Saves probe file from the channel information of the given recording
    extractor

    Parameters
    ----------
    recording: RecordingExtractor
        The recording extractor to save probe file from
    probe_file: str
        file name of .prb or .csv file to save probe information to
    format: str (optional)
        Format for .prb file. It can be either 'klusta' or 'spyking_circus'. Default is None.
    '''
    probe_file = Path(probe_file)
    if not probe_file.parent.is_dir():
        probe_file.parent.mkdir()

    if probe_file.suffix == '.csv':
        # write csv probe file
        with probe_file.open('w') as f:
            if 'location' in recording.get_channel_property_names():
                for chan in recording.get_channel_ids():
                    loc = recording.get_channel_property(chan, 'location')
                    if len(loc) == 2:
                        f.write(str(loc[0]))
                        f.write(',')
                        f.write(str(loc[1]))
                        f.write('\n')
                    elif len(loc) == 3:
                        f.write(str(loc[0]))
                        f.write(',')
                        f.write(str(loc[1]))
                        f.write(',')
                        f.write(str(loc[2]))
                        f.write('\n')
            else:
                raise AttributeError("Recording extractor needs to have "
                                     "'location' property to save .csv probe file")
    elif probe_file.suffix == '.prb':
        _export_prb_file(recording, probe_file, format, radius=radius, dimensions=dimensions)
    elif probe_file.suffix == '.npy':  # YASS foramt
        _export_prb_file(recording, probe_file, format)
    else:
        raise NotImplementedError("Only .csv and .prb probe files can be saved.")


def writeBinaryDatFormat(recording, save_path, transpose=False, dtype='float32'):
    '''Saves the traces of a recording extractor in binary .dat format.

    Parameters
    ----------
    recording: RecordingExtractor
        The recording extractor object to be saved in .dat format
    save_path: str
        The path to the file.
    transpose: bool
        If True the data are transpose (spyking_circus). Default is False (klusta, kilosort, yass)
    dtype: dtype
        Type of the saved data. Default float32

    Returns
    -------

    '''
    save_path = Path(save_path)
    if not transpose:
        if not save_path.suffix == '.dat':
            save_path = save_path.parent / (save_path.name + '.dat')
        with save_path.open('wb') as f:
            np.transpose(np.array(recording.get_traces(), dtype=dtype)).tofile(f)
    elif transpose:
        if not save_path.suffix == '.dat':
            save_path = save_path.parent / (save_path.name + '.dat')
        with save_path.open('wb') as f:
            np.array(recording.get_traces(), dtype=dtype, order='F').tofile(f)
    return save_path


def getSubExtractorsByProperty(extractor, property_name):
    '''Divides Recording or Sorting Extractor based on the property_name (e.g. group)

    Parameters
    ----------
    extractor: RecordingExtractor or SortingExtractor
        The extractor to be subdivided in subextractors
    property_name: str
        The property used to subdivide the extractor

    Returns
    -------
    List of subextractors

    '''
    if isinstance(extractor, RecordingExtractor):
        if property_name not in extractor.get_channel_property_names():
            raise ValueError("'property_name' must be must be a property of the recording channels")
        else:
            sub_list = []
            recording = extractor
            properties = np.array([recording.get_channel_property(chan, property_name)
                                   for chan in recording.get_channel_ids()])
            prop_list = np.unique(properties)
            for prop in prop_list:
                prop_idx = np.where(prop == properties)
                chan_idx = list(np.array(recording.get_channel_ids())[prop_idx])
                sub_list.append(SubRecordingExtractor(recording, channel_ids=chan_idx))
            return sub_list
    elif isinstance(extractor, SortingExtractor):
        if property_name not in extractor.get_unit_propertyNames():
            raise ValueError("'property_name' must be must be a property of the units")
        else:
            sub_list = []
            sorting = extractor
            properties = np.array([sorting.get_unit_property(unit, property_name)
                                   for unit in sorting.get_unit_ids()])
            prop_list = np.unique(properties)
            for prop in prop_list:
                prop_idx = np.where(prop == properties)
                unit_idx = list(np.array(sorting.get_unit_ids())[prop_idx])
                sub_list.append(SubSortingExtractor(sorting, unit_ids=unit_idx))
            return sub_list
    else:
        raise ValueError("'extractor' must be a RecordingExtractor or a SortingExtractor")


def _export_prb_file(recording, file_name, format=None, adjacency_distance=None, graph=False, geometry=True, radius=100,
                     dimensions='all'):
    '''Exports .prb file

    Parameters
    ----------
    recording: RecordingExtractor
        The recording extractor to save probe file from
    file_name: str
        probe filename to be exported to
    format: str
        'klusta' | 'spiking_circus' | 'yass' (defualt=None)
    adjacency_distance: float
        distance to consider 2 channels adjacent (if 'location' is a property)
    graph: bool
        if True graph information is extracted and saved
    geometry:
        if True geometry is saved
    radius: float
        radius for template-matching (if format is 'spyking_circus')
    '''
    if format == 'klusta':
        graph = True
        geometry = False
    elif format == 'spyking_circus':
        graph = False
        geometry = True
    elif format == 'yass':
        graph = False
        geometry = True
    else:
        graph = True
        geometry = True

    file_name = Path(file_name)
    assert file_name is not None
    abspath = file_name.absolute()

    if geometry:
        if 'location' in recording.get_channel_property_names():
            positions = np.array([recording.get_channel_property(chan, 'location')
                                  for chan in recording.get_channel_ids()])
            if dimensions is not None and dimensions is not 'all':
                positions = positions[:, dimensions]
        else:
            print("'location' property is not available and it will not be saved.")
            positions = None
            geometry = False
    else:
        positions = None

    if 'group' in recording.get_channel_property_names():
        groups = np.array([recording.get_channel_property(chan, 'group') for chan in recording.get_channel_ids()])
        channel_groups = np.unique([groups])
    else:
        print("'group' property is not available and it will not be saved.")
        channel_groups = [0]
        groups = np.array([0] * recording.get_num_channels())

    n_elec = recording.get_num_channels()

    # find adjacency graph
    if graph:
        if positions is not None and adjacency_distance is not None:
            adj_graph = []
            for chg in channel_groups:
                group_graph = []
                elecs = list(np.where(groups == chg)[0])
                for i in range(len(elecs)):
                    for j in range(i, len(elecs)):
                        if elecs[i] != elecs[j]:
                            if np.linalg.norm(positions[elecs[i]] - positions[elecs[j]]) < adjacency_distance:
                                group_graph.append((elecs[i], elecs[j]))
                adj_graph.append(group_graph)
        else:
            # all connected by group
            adj_graph = []
            for chg in channel_groups:
                group_graph = []
                elecs = list(np.where(groups == chg)[0])
                for i in range(len(elecs)):
                    for j in range(i, len(elecs)):
                        if elecs[i] != elecs[j]:
                            group_graph.append((elecs[i], elecs[j]))
                adj_graph.append(group_graph)

    if format == 'yass':
        np.save(str(file_name), positions)
        return

    with abspath.open('w') as f:
        f.write('\n')
        if format == 'spyking_circus':
            f.write('total_nb_channels = ' + str(n_elec) + '\n')
            f.write('radius = ' + str(radius) + '\n')
        f.write('channel_groups = {\n')
        if len(channel_groups) > 0:
            for i_chg, chg in enumerate(channel_groups):
                f.write("     " + str(int(chg)) + ": ")
                elecs = list(np.where(groups == chg)[0])
                f.write("\n        {\n")
                f.write("           'channels': " + str(elecs) + ',\n')
                if graph:
                    if len(adj_graph) == 1:
                        f.write("           'graph':  " + str(adj_graph[0]) + ',\n')
                    else:
                        f.write("           'graph':  " + str(adj_graph[i_chg]) + ',\n')
                else:
                    f.write("           'graph':  [],\n")
                if geometry:
                    f.write("           'geometry':  {\n")
                    for i, pos in enumerate(positions[elecs]):
                        f.write('               ' + str(elecs[i]) + ': ' + str(list(pos)) + ',\n')
                    f.write('           }\n')
                f.write('       },\n')
            f.write('}\n')
        else:
            for elec in range(n_elec):
                f.write('    ' + str(elec) + ': ')
                f.write("\n        {\n")
                f.write("           'channels': [" + str(elec) + '],\n')
                f.write("           'graph':  [],\n")
                f.write('        },\n')
            f.write('}\n')



from .spikeforest_views.currentstateview import CurrentStateView
from .spikeforest_views.recordingtableview import RecordingTableView, RecordingSelectComponent
from .spikeforest_views.aggregatedsortingresultstableview import AggregatedSortingResultsTableView
from .recording_views.electrodegeometryview import ElectrodeGeometryView
from .recording_views.timeseriesview import TimeseriesView
from .recording_views.templatesview import TemplatesView
from .recording_views.recordingsummaryview import RecordingSummaryView
from .recording_views.unitstableview import UnitsTableView
from .recording_views.sortingresultstableview import SortingResultsTableView, SortingResultSelectComponent
from .recording_views.sortingresultdetailview import SortingResultDetailView
from .recording_views.featurespaceview import FeatureSpaceView
from .recording_views.clusterview import ClusterView

import vdomr as vd
from mountaintools import client as mt
import json


def get_spikeforest_view_launchers(context):
    launchers = []
    groups = []
    ret = dict(
        groups=groups,
        launchers=launchers
    )

    # General
    groups.append(dict(name='general', label=''))

    launchers.append(dict(
        group='general', name='recording-table', label='Recording table',
        view_class=RecordingTableView,
        context=context, opts=dict(),
        enabled=True
    ))
    launchers.append(dict(
        group='general', name='current-state', label='Current state',
        view_class=CurrentStateView,
        context=context, opts=dict(),
        enabled=True
    ))
    # MEDIUM TODO: this should be a component rather than a launcher
    launchers.append(dict(
        group='general', name='recording-select',
        component_class=RecordingSelectComponent,
        context=context, opts=dict(),
        enabled=True
    ))

    recording_context = context.recordingContext(context.currentRecordingId())

    # Aggregated sorting results
    if context.hasAggregatedSortingResults():
        groups.append(dict(name='aggregated_sorting_results', label='Aggregated results'))

        launchers.append(dict(
            group='aggregated_sorting_results', name='aggregated-results-table', label='Results table',
            view_class=AggregatedSortingResultsTableView,
            context=context, opts=dict(),
            always_open_new=False,
            enabled=True
        ))

    # Recording
    if recording_context:
        groups.append(dict(name='recording', label='Recording', sublabel=context.currentRecordingId()))

        launchers.append(dict(
            group='recording', name='recording-summary', label='Recording summary',
            view_class=RecordingSummaryView,
            context=recording_context, opts=dict(),
            always_open_new=True,
            enabled=(recording_context is not None)
        ))
        launchers.append(dict(
            group='recording', name='electrode-geometry', label='Electrode geometry',
            view_class=ElectrodeGeometryView,
            context=recording_context, opts=dict(),
            enabled=(recording_context is not None)
        ))
        launchers.append(dict(
            group='recording', name='timeseries', label='Timeseries',
            view_class=TimeseriesView,
            context=recording_context, opts=dict(),
            always_open_new=True,
            enabled=(recording_context is not None)
        ))
        if recording_context.hasIntraRecording():
            launchers.append(dict(
                group='recording', name='intra-timeseries', label='Intra-timeseries',
                view_class=TimeseriesView,
                context=recording_context.intraRecordingContext(),
                always_open_new=True,
                enabled=(recording_context is not None)
            ))

    # True sorting
    if recording_context and recording_context.trueSortingContext():
        true_sorting_context = recording_context.trueSortingContext()

        groups.append(dict(name='true-sorting', label='True sorting'))
        launchers.append(dict(
            group='true-sorting', name='true-templates', label='Templates',
            view_class=TemplatesView,
            context=true_sorting_context, opts=dict(),
            always_open_new=True,
            enabled=(true_sorting_context is not None)
        ))
        launchers.append(dict(
            group='true-sorting', name='true-units-info', label='Units info',
            view_class=UnitsTableView,
            context=true_sorting_context, opts=dict(),
            always_open_new=True,
            enabled=(true_sorting_context is not None)
        ))
        launchers.append(dict(
            group='true-sorting', name='feature-space', label='Feature space',
            view_class=FeatureSpaceView,
            context=true_sorting_context, opts=dict(),
            always_open_new=True,
            enabled=(len(true_sorting_context.selectedUnitIds()) > 0)
        ))
        launchers.append(dict(
            group='true-sorting', name='clusters', label='Clusters',
            view_class=ClusterView,
            context=true_sorting_context, opts=dict(),
            always_open_new=True,
            enabled=(len(true_sorting_context.selectedUnitIds()) > 0)
        ))

        dict(name='unit', label='Unit')
        launchers.append(dict(
            group='true-sorting', name='test', label='Test',
            view_class=TemplatesView,
            context=true_sorting_context, opts=dict(),
            always_open_new=True,
            enabled=(true_sorting_context.currentUnitId() is not None)
        ))

    # Sorting results
    if recording_context and (len(recording_context.sortingResultNames()) > 0):
        groups.append(dict(name='sorting-results', label='Sorting results'))
        launchers.append(dict(
            group='sorting-results', name='sorting-results-table', label='Sorting results table',
            view_class=SortingResultsTableView,
            context=recording_context, opts=dict(),
            always_open_new=True,
            enabled=(len(recording_context.sortingResultNames()) > 0)
        ))
        launchers.append(dict(
            group='sorting-results', name='sorting-result-select',
            component_class=SortingResultSelectComponent,
            context=recording_context, opts=dict(),
            always_open_new=True,
            enabled=(len(recording_context.sortingResultNames()) > 0)
        ))

    # Sorting result
    if recording_context and recording_context.currentSortingResult():
        srname = recording_context.currentSortingResult()
        sorting_result_context = recording_context.sortingResultContext(srname)

        groups.append(dict(name='sorting-result', label='Sorting result', sublabel=srname))
        launchers.append(dict(
            group='sorting-result', name='sorting-result-details', label='Details',
            view_class=SortingResultDetailView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(sorting_result_context is not None)
        ))
        launchers.append(dict(
            group='sorting-result', name='templates', label='Templates',
            view_class=TemplatesView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(sorting_result_context is not None)
        ))
        launchers.append(dict(
            group='sorting-result', name='units-info', label='Units info',
            view_class=UnitsTableView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(sorting_result_context is not None)
        ))
        launchers.append(dict(
            group='sorting-result', name='feature-space', label='Feature space',
            view_class=FeatureSpaceView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(len(sorting_result_context.selectedUnitIds()) > 0)
        ))
        launchers.append(dict(
            group='sorting-result', name='clusters', label='Clusters',
            view_class=ClusterView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(len(sorting_result_context.selectedUnitIds()) > 0)
        ))
        launchers.append(dict(
            group='sorting-result', name='console-out', label='Console output',
            view_class=ConsoleOutView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(sorting_result_context.consoleOutputPath() is not None)
        ))
        launchers.append(dict(
            group='sorting-result', name='exec-stats', label='Execution stats',
            view_class=ExecutionStatsView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(sorting_result_context.executionStats() is not None)
        ))
        launchers.append(dict(
            group='sorting-result', name='comparison-with-truth', label='Comparison with truth',
            view_class=ComparisonWithTruthView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(sorting_result_context.comparisonWithTruthPath() is not None)
        ))

        dict(name='unit', label='Unit')
        launchers.append(dict(
            group='sorting-result', name='test', label='Test',
            view_class=TemplatesView,
            context=sorting_result_context, opts=dict(),
            always_open_new=True,
            enabled=(sorting_result_context.currentUnitId() is not None)
        ))

    return ret


class ConsoleOutView(vd.Component):
    def __init__(self, *, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        if not context.consoleOutputPath():
            self._text = 'no console output found'
        else:
            self._text = mt.loadText(path=context.consoleOutputPath()) or 'unable to load console output'

    def setSize(self, size):
        if self._size != size:
            self._size = size

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Console out'

    def render(self):
        return vd.components.ScrollArea(vd.pre(self._text), height=self._size[1])


class ExecutionStatsView(vd.Component):
    def __init__(self, *, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._stats = context.executionStats()

    def setSize(self, size):
        if self._size != size:
            self._size = size

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Exec stats'

    def render(self):
        if not self._stats:
            return vd.div('No stats found')
        return vd.div(vd.pre(json.dumps(self._stats, indent=4)))


class ComparisonWithTruthView(vd.Component):
    def __init__(self, *, context, opts=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        if not context.comparisonWithTruthPath():
            self._object = None
        else:
            self._object = mt.loadObject(path=context.comparisonWithTruthPath())

    def setSize(self, size):
        if self._size != size:
            self._size = size

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Comparison with truth'

    def render(self):
        if not self._object:
            return vd.div('Unable to load comparison data.')
        return vd.components.ScrollArea(vd.pre(json.dumps(self._object, indent=4)), height=self._size[1])



import vdomr as vd
import time
import sys
import uuid
import json
import os
import uuid

source_path = os.path.dirname(os.path.realpath(__file__))


class ElectrodeGeometryView(vd.Component):
    def __init__(self, context, opts=None, prepare_result=None):
        vd.Component.__init__(self)
        self._context = context
        self._size = None
        self._widget = ElectrodeGeometryWidget(recording=context.recordingExtractor())
        self.setSize((100, 100))
        self._update_state()
        self._widget.onStateChanged(self._handle_state_changed)
        self._context.onCurrentChannelChanged(self._update_state)

    @staticmethod
    def prepareView(context, opts):
        context.initialize()

    def setSize(self, size):
        if self._size == size:
            return
        self._size = size
        self._widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Electrode geometry'

    def title(self):
        return 'Electrode geometry for {}'.format(self._context.recordingLabel())

    def render(self):
        return vd.div(self._widget)

    def _handle_state_changed(self):
        self._context.setCurrentChannel(self._widget.currentChannel())

    def _update_state(self):
        self._widget.setCurrentChannel(self._context.currentChannel())


class ElectrodeGeometryWidget(vd.Component):
    def __init__(self, recording):
        vd.Component.__init__(self)
        self._recording = recording
        self._div_id = 'electrode-geometry-' + str(uuid.uuid4())
        self._size = (200, 200)
        self._current_channel = None
        self._state_changed_handlers = []

        # MEDIUM TODO clean up these rel .js reference paths throughout
        vd.devel.loadJavascript(path=source_path + '/../../spikeforestwidgets/templatewidget/canvaswidget.js')
        vd.devel.loadJavascript(path=source_path + '/electrodegeometrywidget.js')
        vd.devel.loadJavascript(path=source_path + '/../..//spikeforestwidgets/dist/jquery-3.3.1.min.js')

    def setSize(self, size):
        if self._size == size:
            return
        self._size = size
        self._update_widget()

    def currentChannel(self):
        return self._current_channel

    def setCurrentChannel(self, ch):
        if self._current_channel == ch:
            return
        self._current_channel = ch
        self._update_widget()
        for handler in self._state_changed_handlers:
            handler()

    def onStateChanged(self, handler):
        self._state_changed_handlers.append(handler)

    def _on_state_change(self, current_electrode_index):
        current_electrode_index = int(current_electrode_index)
        channel_ids = self._recording.get_channel_ids()
        if current_electrode_index >= 0:
            ch = channel_ids[int(current_electrode_index)]
        else:
            ch = None
        self.setCurrentChannel(ch)

    def _update_widget(self):
        channel_ids = self._recording.get_channel_ids()
        locations = self._recording.get_channel_locations(channel_ids=channel_ids)
        locations = [[loc[0], loc[1]] for loc in locations]
        try:
            current_electrode_index = channel_ids.index(self._current_channel)
        except:
            current_electrode_index = -1

        js = """
        let W = (window.geom_widgets||{})['{component_id}'];
        W.setSize({width},{height});
        W.setElectrodeLocations({locations});
        W.setElectrodeLabels({labels});
        W.setCurrentElectrodeIndex({current_electrode_index});

        let elmt = $('#{div_id}');
        elmt.css({width:'{width}px',height:'{height}px'})
        """
        js = js.replace('{div_id}', self._div_id)
        js = js.replace('{component_id}', self.componentId())
        js = js.replace('{locations}', json.dumps(locations))
        js = js.replace('{labels}', json.dumps([ch for ch in channel_ids]))
        js = js.replace('{current_electrode_index}', str(current_electrode_index))
        js = js.replace('{width}', str(self._size[0]))
        js = js.replace('{height}', str(self._size[1]))
        self.executeJavascript(js)

    def render(self):
        div = vd.div(id=self._div_id)
        return div

    def postRenderScript(self):
        state_change_callback_id = 'state-change-callback-' + str(uuid.uuid4())
        vd.register_callback(state_change_callback_id, lambda current_electrode_index: self._on_state_change(current_electrode_index=current_electrode_index))

        js = """
        let W=new window.ElectrodeGeometryWidget();
        if (!window.geom_widgets) window.geom_widgets={};
        window.geom_widgets['{component_id}']=W;
        W.onCurrentElectrodeIndexChanged(function() {
            window.vdomr_invokeFunction('{state_change_callback_id}', [Number(W.currentElectrodeIndex())], {})
        });
        let elmt = $('#{div_id}');
        elmt.append(W.element());
        """
        js = js.replace('{div_id}', self._div_id)
        js = js.replace('{component_id}', self.componentId())

        js = js.replace('{state_change_callback_id}', str(state_change_callback_id))
        self._update_widget()
        return js



import uuid
from spikeforest import mdaio
import io
import base64
import vdomr as vd
import os
import numpy as np
import mlprocessors as mlpr
from spikeforest import SFMdaRecordingExtractor
from mountaintools import client as mt
import spikeextractors as se
from spikeforest import EfficientAccessRecordingExtractor
import mtlogging
import time

source_path = os.path.dirname(os.path.realpath(__file__))


def _mda32_to_base64(X):
    f = io.BytesIO()
    mdaio.writemda32(X, f)
    return base64.b64encode(f.getvalue()).decode('utf-8')


class TimeseriesWidget(vd.Component):
    def __init__(self, *, recording, sorting=None, unit_ids=None, start_frame=0, end_frame=None, size=(800, 400)):
        vd.Component.__init__(self)

        vd.devel.loadBootstrap()
        vd.devel.loadCss(url='https://stackpath.bootstrapcdn.com/font-awesome/4.7.0/css/font-awesome.min.css')
        vd.devel.loadJavascript(path=source_path + '/mda.js')
        vd.devel.loadJavascript(path=source_path + '/timeseriesmodel.js')
        vd.devel.loadJavascript(path=source_path + '/canvaswidget.js')
        vd.devel.loadJavascript(path=source_path + '/timeserieswidget.js')
        vd.devel.loadJavascript(path=source_path + '/../dist/jquery-3.3.1.min.js')

        self._div_id = 'SFTimeseriesWidget-' + str(uuid.uuid4())
        self._recording = recording
        self._multiscale_recordings = _create_multiscale_recordings(recording=recording, progressive_ds_factor=3)
        self._segment_size = 1000
        self._data_segments_set = dict()

        if sorting:
            self._sorting = sorting
            if not unit_ids:
                unit_ids = sorting.get_unit_ids()
            # This is not ideal as it seems possible to get this information directly from the recording
            # Alas we cannot be sure this recording (as opposed to it's parent) was the one used for sorting
            if not end_frame:
                end_frame = len(recording.get_num_frames())
            # spike_trains_list = ['[{}]'.format(','.join(str(x) for x in
            #                                             sorting.get_unit_spike_train(u, start_frame=start_frame, end_frame=end_frame)))
            #                      for u in unit_ids]
            spike_trains_str = '[' + ','.join(spike_trains_str) + ']'
        else:
            spike_trains_str = '[[]]'

        # js_lines=[
        #         "window.sfdata=window.sfdata||{}",
        #         "window.sfdata.test=1",
        #         "window.sfdata['array_b64_{div_id}']='{b64}'".format(div_id=self._div_id,
        #             b64=self._array_b64),
        #         "window.sfdata['spike_times']={}".format(spike_trains_str)
        #             ]
        # js = ";".join(js_lines)
        # vd.devel.loadJavascript(js=js)

        # important to set some data here so that the auto-scaling can take place
        self._set_data_segment(ds_factor=1, segment_num=0, autoscale=True)

        self._size = size

    @mtlogging.log()
    def _set_data_segment(self, *, ds_factor, segment_num, autoscale=False):
        code0 = '{}-{}'.format(int(ds_factor), int(segment_num))
        if code0 in self._data_segments_set:
            return

        ds_factor = int(ds_factor)
        self._data_segments_set[code0] = True

        # X = _compute_data_segment(recording=self._recording, ds_factor=ds_factor, segment_num=segment_num, segment_size=self._segment_size)
        if ds_factor == 1:
            X = _extract_data_segment(recording=self._recording, segment_num=segment_num, segment_size=self._segment_size)
        else:
            rx = self._multiscale_recordings[ds_factor]
            X = _extract_data_segment(recording=rx, segment_num=segment_num, segment_size=self._segment_size * 2)

        # X = ComputeDataSegment.execute(
        #     recording=self._recording,
        #     ds_factor=ds_factor,
        #     segment_num=segment_num,
        #     segment_size=self._segment_size,
        #     array_out=True
        # ).outputs['array_out']

        X_b64 = _mda32_to_base64(X)
        js = """
        let TS = window.timeseries_models['{component_id}'];
        let X=new window.Mda();
        X.setFromBase64('{X_b64}');
        TS.setDataSegment({ds_factor}, {segment_num}, X);
        """

        if autoscale:
            js = js + """
            let W = window.timeseries_widgets['{component_id}'];
            W.autoScale();
            """

        js = js.replace('{component_id}', self.componentId())
        js = js.replace('{X_b64}', X_b64)
        js = js.replace('{ds_factor}', str(ds_factor))
        js = js.replace('{segment_num}', str(segment_num))

        self.executeJavascript(js=js)

    def setSize(self, size):
        if self._size == size:
            return
        self._size = size
        self._update_size()

    def size(self):
        return self._size

    def _update_size(self):
        js = """
        let W=window.timeseries_widgets['{component_id}'];
        W.setSize({width},{height});
        $('#{div_id}').css({width:'{width}px',height:'{height}px'});
        """
        js = js.replace('{div_id}', self._div_id)
        js = js.replace('{component_id}', self.componentId())
        js = js.replace('{width}', str(self._size[0]))
        js = js.replace('{height}', str(self._size[1]))
        self.executeJavascript(js)

    def render(self):
        div = vd.div(id=self._div_id)
        self._update_size()
        return div

    def postRenderScript(self):
        js = """
        if (!window.timeseries_models) window.timeseries_models={};
        if (!window.timeseries_models['{component_id}']) {
            let TS0=new window.TimeseriesModel({samplerate:{samplerate}, num_channels:{num_channels}, num_timepoints:{num_timepoints}, segment_size:{segment_size}});
            window.timeseries_models['{component_id}'] = TS0;
        }
        let TS = window.timeseries_models['{component_id}'];

        TS.onRequestDataSegment(request_data_segment);
        
        if (!window.spike_times) window.spike_times={};
        //window.spike_times['{component_id}']={spike_trains}";

        function request_data_segment(ds_factor, num) {
            window.vdomr_invokeFunction('{request_data_segment_callback_id}', [ds_factor, num], {});
        }

        if (!window.timeseries_widgets) window.timeseries_widgets={};
        if (!window.timeseries_widgets['{component_id}']) {
            let W0=new window.TimeseriesWidget();
            W0.setSyncGroup('test');
            W0.setTimeseriesModel(TS);
            window.timeseries_widgets['{component_id}'] = W0;
        }
        let W=window.timeseries_widgets['{component_id}'];
        //W.setMarkers(window.spike_times['{component_id}']);
        $('#{div_id}').empty();
        $('#{div_id}').append(W.element());
        """

        request_data_segment_callback_id = 'request-data-segment-' + str(uuid.uuid4())
        vd.register_callback(request_data_segment_callback_id, lambda ds_factor, segment_num: self._set_data_segment(ds_factor=ds_factor, segment_num=segment_num))

        js = js.replace('{div_id}', self._div_id)
        js = js.replace('{component_id}', self.componentId())
        # js = js.replace('{spike_trains}', spike_trains_str)
        js = js.replace('{samplerate}', str(self._recording.get_sampling_frequency()))
        js = js.replace('{num_channels}', str(self._recording.get_num_channels()))
        js = js.replace('{num_timepoints}', str(self._recording.get_num_frames()))
        js = js.replace('{segment_size}', str(self._segment_size))
        js = js.replace('{request_data_segment_callback_id}', request_data_segment_callback_id)
        return js


def _do_downsample(X, ds_factor):
    M = X.shape[0]
    N = X.shape[1]
    N2 = N // ds_factor
    X_reshaped = np.reshape(X, (M, N2, ds_factor), order='C')
    ret = np.zeros((M, N2 * 2))
    ret[:, 0::2] = np.min(X_reshaped, axis=2)
    ret[:, 1::2] = np.max(X_reshaped, axis=2)
    return ret


@mtlogging.log()
def _extract_data_segment(*, recording, segment_num, segment_size):
    segment_num = int(segment_num)
    segment_size = int(segment_size)
    a1 = segment_size * segment_num
    a2 = segment_size * (segment_num + 1)
    if a2 > recording.get_num_frames():
        a2 = recording.get_num_frames()
    X = recording.get_traces(start_frame=a1, end_frame=a2)
    return X

# to remove
# class ComputeDataSegment(mlpr.Processor):
#     recording = mlpr.Input()
#     segment_size = mlpr.IntegerParameter()
#     segment_num = mlpr.IntegerParameter()
#     ds_factor = mlpr.IntegerParameter()
#     array_out = mlpr.OutputArray()

#     def run(self):
#         X = _compute_data_segment(
#             recording=self.recording,
#             segment_size=self.segment_size,
#             segment_num=self.segment_num,
#             ds_factor=int(self.ds_factor), # not sure why we need to cast as an int here
#         )
#         print('===================== Saving file to:', self.array_out, X.shape)
#         np.save(self.array_out, X)


def precomputeMultiscaleRecordings(*, recording):
    _create_multiscale_recordings(recording=recording, progressive_ds_factor=3)


def _create_multiscale_recordings(*, recording, progressive_ds_factor):
    ret = dict()
    current_rx = recording
    current_ds_factor = 1
    N = recording.get_num_frames()
    recording_has_minmax = False
    while current_ds_factor * progressive_ds_factor < N:
        current_rx = _DownsampledRecordingExtractor(recording=current_rx, ds_factor=progressive_ds_factor, input_has_minmax=recording_has_minmax)
        current_rx = EfficientAccessRecordingExtractor(recording=current_rx)
        current_ds_factor = current_ds_factor * progressive_ds_factor
        ret[current_ds_factor] = current_rx
        recording_has_minmax = True
    return ret


class _DownsampledRecordingExtractor(se.RecordingExtractor):
    def __init__(self, *, recording, ds_factor, input_has_minmax):
        se.RecordingExtractor.__init__(self)
        self._recording = recording
        self._ds_factor = ds_factor
        self._input_has_minmax = input_has_minmax
        self.copy_channel_properties(recording)

    def hash(self):
        return mt.sha1OfObject(dict(
            name='downsampled-recording-extractor',
            version=2,
            recording=self._recording.hash(),
            ds_factor=self._ds_factor,
            input_has_minmax=self._input_has_minmax
        ))

    def get_channel_ids(self):
        # same channel IDs
        return self._recording.get_channel_ids()

    def get_num_frames(self):
        if self._input_has_minmax:
            # number of frames is just /ds_factor (but not quite -- tricky!)
            return ((self._recording.get_num_frames() // 2) // self._ds_factor) * 2
        else:
            # need to double because we will now keep track of mins and maxs
            return (self._recording.get_num_frames() // self._ds_factor) * 2

    def get_sampling_frequency(self):
        if self._input_has_minmax:
            # sampling frequency is just /ds_factor
            return self._recording.get_sampling_frequency() / self._ds_factor
        else:
            # need to double because we will now keep track of mins and maxes
            return (self._recording.get_sampling_frequency() / self._ds_factor) * 2

    def get_traces(self, channel_ids=None, start_frame=None, end_frame=None):
        ds_factor = self._ds_factor
        if self._input_has_minmax:
            # get the traces *ds_factor
            X = self._recording.get_traces(
                channel_ids=channel_ids,
                start_frame=start_frame * ds_factor,
                end_frame=end_frame * ds_factor
            )
            X_mins = X[:, 0::2]  # here are the minimums
            X_maxs = X[:, 1::2]  # here are the maximums
            X_mins_reshaped = np.reshape(X_mins, (X_mins.shape[0], X_mins.shape[1] // ds_factor, ds_factor), order='C')
            X_maxs_reshaped = np.reshape(X_maxs, (X_maxs.shape[0], X_maxs.shape[1] // ds_factor, ds_factor), order='C')
            # the size of the output is the size X divided by ds_factor
            ret = np.zeros((X.shape[0], X.shape[1] // ds_factor))
            ret[:, 0::2] = np.min(X_mins_reshaped, axis=2)  # here are the mins
            ret[:, 1::2] = np.max(X_maxs_reshaped, axis=2)  # here are the maxs
            return ret
        else:
            X = self._recording.get_traces(
                channel_ids=channel_ids,
                start_frame=start_frame * self._ds_factor // 2,
                end_frame=end_frame * self._ds_factor // 2
            )
            X_reshaped = np.reshape(X, (X.shape[0], X.shape[1] // ds_factor, ds_factor), order='C')
            ret = np.zeros((X.shape[0], (X.shape[1] // ds_factor) * 2))
            ret[:, 0::2] = np.min(X_reshaped, axis=2)
            ret[:, 1::2] = np.max(X_reshaped, axis=2)
            return ret

    @staticmethod
    def write_recording(recording, save_path):
        EfficientAccessRecordingExtractor(recording=recording, _dest_path=save_path)



from matplotlib import pyplot as plt
import numpy as np
from matplotlib.ticker import MaxNLocator

def plot_timeseries(*, recording, sorting=None, channels=None, trange=None, width=None, height=None):
    W = TimeseriesWidget(
        recording=recording,
        sorting=sorting,
        channels=channels,
        trange=trange,
        width=width,
        height=height
    )
    W.plot()


def view_timeseries(*, recording, sorting=None, channels=None, trange=None, width=None, height=None):
    W = TimeseriesWidget(
        recording=recording,
        sorting=sorting,
        channels=channels,
        trange=trange,
        width=width,
        height=height
    )
    W.display()


class TimeseriesWidget:
    def __init__(self, *, recording, sorting=None, channels=None, trange=None, width=None, height=None):
        self._recording = recording
        self._sorting = sorting
        self._samplerate = recording.get_sampling_frequency()
        self._width = width
        self._height = height
        self._visible_channels = channels
        if self._width is None:
            self._width = 12
        if self._height is None:
            self._height = 6
        if self._visible_channels is None:
            self._visible_channels = recording.get_channel_ids()
        self._visible_trange = trange
        if self._visible_trange is None:
            self._visible_trange = [0, np.minimum(10000, recording.get_num_frames())]
        self._initialize_stats()
        self._vspacing = self._mean_channel_std * 15
        self._visible_trange = self._fix_trange(self._visible_trange)
        # self._update_plot()

    def _init_widgets(self):
        import ipywidgets as widgets
        self._widget = widgets.Output()
        self._control_panel = self._create_control_panel()
        self._main_widget = widgets.VBox([self._control_panel, self._widget])

    def plot(self):
        self._do_plot()

    def display(self):
        self._init_widgets()
        self._update_plot()
        display(self._main_widget)

    def widget(self):
        return self._widget

    def figure(self):
        return self._figure

    def _update_plot(self):
        self._widget.clear_output(wait=True)
        with self._widget:
            self._do_plot()

    def _do_plot(self):
        chunk0 = self._recording.get_traces(
            channel_ids=self._visible_channels,
            start_frame=self._visible_trange[0],
            end_frame=self._visible_trange[1]
        )
        plt.xlim(self._visible_trange[0] / self._samplerate, self._visible_trange[1] / self._samplerate)
        plt.ylim(-self._vspacing, self._vspacing * len(self._visible_channels))
        plt.gcf().set_size_inches(self._width, self._height)
        plt.gca().get_xaxis().set_major_locator(MaxNLocator(prune='both'))
        plt.gca().get_yaxis().set_ticks([])
        plt.xlabel('Time (sec)')

        self._plots = {}
        self._plot_offsets = {}
        offset0 = self._vspacing * (len(self._visible_channels) - 1)
        tt = np.arange(self._visible_trange[0], self._visible_trange[1]) / self._samplerate
        for im, m in enumerate(self._visible_channels):
            self._plot_offsets[m] = offset0
            self._plots[m] = plt.plot(tt, self._plot_offsets[m] + chunk0[im, :])
            offset0 = offset0 - self._vspacing
        self._figure = plt.gcf()
        # plt.show()

    def _pan_left(self):
        self._pan(-0.1)

    def _pan_right(self):
        self._pan(0.1)

    def _pan(self, factor):
        span = self._visible_trange[1] - self._visible_trange[0]
        delta = int(span * factor)
        new_trange = [self._visible_trange[0] + delta, self._visible_trange[1] + delta]
        new_trange = self._fix_trange(new_trange)
        self._visible_trange = new_trange
        self._update_plot()

    def _fix_trange(self, trange):
        N = self._recording.get_num_frames()
        if trange[1] > N:
            trange[0] += N - trange[1]
            trange[1] += N - trange[1]
        if trange[0] < 0:
            trange[1] += -trange[0]
            trange[0] = 0
        trange[0] = np.maximum(0, trange[0])
        trange[1] = np.minimum(N, trange[1])
        return trange

    def _scale_up(self):
        self._scale(1.2)

    def _scale_down(self):
        self._scale(1 / 1.2)

    def _scale(self, factor):
        self._vspacing /= factor
        self._update_plot()

    def _zoom_in(self):
        self._zoom(1.2)

    def _zoom_out(self):
        self._zoom(1 / 1.2)

    def _zoom(self, factor):
        span = self._visible_trange[1] - self._visible_trange[0]
        new_span = int(np.maximum(30, span / factor))
        tcenter = int((self._visible_trange[0] + self._visible_trange[1]) / 2)
        new_trange = [int(tcenter - new_span / 2), int(tcenter - new_span / 2 + new_span - 1)]
        new_trange = self._fix_trange(new_trange)
        self._visible_trange = new_trange
        self._update_plot()

    def _initialize_stats(self):
        self._channel_stats = {}
        # M=self._reader.numChannels()
        # N=self._reader.numTimepoints()
        chunk0 = self._recording.get_traces(
            channel_ids=self._visible_channels,
            start_frame=self._visible_trange[0],
            end_frame=self._visible_trange[1]
        )
        # chunk0=self._reader.getChunk(channels=self._visible_channels,trange=self._visible_trange)
        M0 = chunk0.shape[0]
        N0 = chunk0.shape[1]
        for ii in range(M0):
            self._channel_stats[self._visible_channels[ii]] = self._compute_channel_stats_from_data(chunk0[ii, :])
        self._mean_channel_std = np.mean([self._channel_stats[m]['std'] for m in self._visible_channels])

    def _compute_channel_stats_from_data(self, X):
        return dict(
            mean=np.mean(X),
            std=np.std(X)
        )

    def _create_control_panel(self):
        def on_zoom_in(b):
            self._zoom_in()

        def on_zoom_out(b):
            self._zoom_out()

        def on_pan_left(b):
            self._pan_left()

        def on_pan_right(b):
            self._pan_right()

        def on_scale_up(b):
            self._scale_up()

        def on_scale_down(b):
            self._scale_down()

        zoom_in = widgets.Button(icon='plus-square', tooltip="Zoom In", layout=dict(width='40px'))
        zoom_in.on_click(on_zoom_in)
        zoom_out = widgets.Button(icon='minus-square', tooltip="Zoom Out", layout=dict(width='40px'))
        zoom_out.on_click(on_zoom_out)
        pan_left = widgets.Button(icon='arrow-left', tooltip="Pan left", layout=dict(width='40px'))
        pan_left.on_click(on_pan_left)
        pan_right = widgets.Button(icon='arrow-right', tooltip="Pan right", layout=dict(width='40px'))
        pan_right.on_click(on_pan_right)
        scale_up = widgets.Button(icon='arrow-up', tooltip="Scale up", layout=dict(width='40px'))
        scale_up.on_click(on_scale_up)
        scale_down = widgets.Button(icon='arrow-down', tooltip="Scale down", layout=dict(width='40px'))
        scale_down.on_click(on_scale_down)
        self._debug = widgets.Output()
        return widgets.HBox([zoom_in, zoom_out, pan_left, pan_right, scale_up, scale_down, self._debug])



# %% Change working directory from the workspace root to the ipynb file location. Turn this addition off with the DataScience.changeDirOnImportExport setting
from mountaintools import client as mt
import sfdata as sf
import spikeforest_analysis as sa
import os
try:
    os.chdir(os.path.join(os.getcwd(), 'working/spikeforest_test1'))
    print(os.getcwd())
except:
    pass

# %%
# path=ca.saveFile('/mnt/home/jjun/src/sf2dev/spikesorters/containers/yass/yass.simg')
# print(path)

# %%
get_ipython().run_line_magic('load_ext', 'autoreload')
get_ipython().run_line_magic('autoreload', '2')


# %%
password = os.environ.get('SPIKEFOREST_PASSWORD', None)
mt.autoConfig(collection='spikeforest', key='spikeforest2-readwrite',
              password=password, ask_password=True)


# %%
os.environ['MLPROCESSORS_FORCE_RUN'] = 'TRUE'
recording = dict(
    directory='kbucket://15734439d8cf/groundtruth/magland_synth/datasets_noise10_K10_C4/001_synth')
recordings = [recording]


# %%
sorter_ms4_thr3 = dict(
    name='MountainSort4-thr3',
    processor_name='MountainSort4',
    params=dict(
        detect_sign=-1,
        adjacency_radius=50,
        detect_threshold=3.1
    )
)

sorter_sc = dict(
    name='SpykingCircus',
    processor_name='SpykingCircus',
    params=dict(
        detect_sign=-1,
        adjacency_radius=50
    )
)

sorter_yass = dict(
    name='yass',
    processor_name='yass',
    params=dict(
        detect_sign=-1,
        adjacency_radius=50
    )
)
# sorters=[sorter_ms4_thr3,sorter_sc,sorter_yass]
# sorters=[sorter_yass]
sorters = [sorter_sc]


# %%
# compute_resource='jfm-laptop'
compute_resource = None
recordings_B = sa.summarize_recordings(
    recordings=recordings, compute_resource=compute_resource)
sorting_results = []
for sorter in sorters:
    sortings_A = sa.sort_recordings(
        sorter=sorter, recordings=recordings_B, compute_resource=compute_resource)
    sortings_B = sa.summarize_sortings(
        sortings=sortings_A, compute_resource=compute_resource)
    sortings_C = sa.compare_sortings_with_truth(
        sortings=sortings_B, compute_resource=compute_resource)
    sorting_results = sorting_results+sortings_C


# %%
sortings_A


# %%
summaries = sa.summarize_recordings(
    recordings=[recording], compute_resource=None)  # 'ccmlin008-default')
recording['summary'] = summaries[0]


# %%
sorter_ms4_thr3 = dict(
    name='MountainSort4-thr3',
    processor_name='MountainSort4',
    params=dict(
        detect_sign=-1,
        adjacency_radius=50,
        detect_threshold=3.1
    )
)

sorter_sc = dict(
    name='SpykingCircus',
    processor_name='SpykingCircus',
    params=dict(
        detect_sign=-1,
        adjacency_radius=50
    )
)

sorter_yass = dict(
    name='yass',
    processor_name='yass',
    params=dict(
        detect_sign=-1,
        adjacency_radius=50
    )

)

compute_resource = None
sorting_ms4 = sa.sort_recordings(
    recordings=[recording], sorter=sorter_ms4_thr3, compute_resource=compute_resource)[0]
sorting_sc = sa.sort_recordings(
    recordings=[recording], sorter=sorter_sc, compute_resource=compute_resource)[0]
# sorting_yass=sa.sort_recordings(recordings=[recording],sorter=sorter_yass,compute_resource=compute_resource)[0]


# %%
display(sorting_ms4)
print(mt.loadText(path=sorting_ms4['console_out'])[0:1000])
display(sorting_sc)
print(mt.loadText(path=sorting_sc['console_out'])[0:1000])


# %%
sorting_ms4['summary'] = sa.summarize_sortings(
    sortings=[sorting_ms4], compute_resource=compute_resource)
sorting_ms4['comparison_with_truth'] = sa.compare_sortings_with_truth(
    sortings=[sorting_ms4], compute_resource=compute_resource)


# %%
sorting_ms4


# %%
mt.loadObject(
    path='sha1://cbc3f0d7beb8f94d3bf4287b38ca4b05782f94ec/table.json')


# %%



from mountaintools import client as mt
import random
import string
import datetime


def clearBatch(*, batch_name, test_one=False):
    batch = mt.loadObject(key=dict(batch_name=batch_name))
    jobs = batch['jobs']

    if test_one and (len(jobs) > 0):
        jobs = [jobs[0]]

    setBatchStatus(
        batch_name=batch_name,
        status='clearing_batch'
    )
    _clear_job_results(jobs=jobs, incomplete_only=False)
    setBatchStatus(
        batch_name=batch_name,
        status='finished_clearing_batch'
    )


def prepareBatch(*, batch_name, test_one=False):
    batch = mt.loadObject(key=dict(batch_name=batch_name))
    jobs = batch['jobs']

    if test_one and (len(jobs) > 0):
        jobs = [jobs[0]]

    setBatchStatus(
        batch_name=batch_name,
        status='preparing_batch'
    )
    _clear_job_results(jobs=jobs, incomplete_only=True)

    setBatchStatus(
        batch_name=batch_name,
        status='downloading_recordings'
    )
    _download_recordings(jobs=jobs)

    setBatchStatus(
        batch_name=batch_name,
        status='finished_preparing_batch'
    )


def runBatch(*, batch_name, test_one=False):
    print('Loading batch object...')
    batch = mt.loadObject(key=dict(batch_name=batch_name))
    jobs = batch['jobs']

    if test_one and (len(jobs) > 0):
        jobs = [jobs[0]]

    print('Running batch with {} jobs...'.format(len(jobs)))
    for job in jobs:
        _run_job(job)


def setBatchStatus(*, batch_name, status):
    mt.setValue(key=dict(name='spikeforest_batch_status', batch_name=batch_name), value=status)


def _do_sort_recording(job):
    try:
        from .sort_recording import sort_recording
    except:
        print('Problem importing sort_recording. You probably need to install one or more python packages.')
        raise
    return sort_recording(sorter=job['sorter'], recording=job['recording'])


def _do_summarize_recording(job):
    try:
        from .summarize_recording import summarize_recording
    except:
        print('Problem importing summarize_recording. You probably need to install one or more python packages.')
        raise
    return summarize_recording(recording=job['recording'])


def _do_run_job(job):
    if job['command'] == 'sort_recording':
        return _do_sort_recording(job)
    elif job['command'] == 'summarize_recording':
        return _do_summarize_recording(job)
    else:
        return dict(error='Invalid job command: ' + job['command'])


def _set_job_status(job, status):
    mt.saveObject(key=dict(name='job_status', job=job), object=status)


def _run_job(job):
    val = mt.getValue(key=job)
    if val:
        return
    code = ''.join(random.choice(string.ascii_uppercase) for x in range(10))
    if not mt.setValue(key=job, value='in-process-' + code, overwrite=False):
        return
    status = dict(
        time_started=_make_timestamp(),
        status='running'
    )
    _set_job_status(job, status)

    print('Running job: ' + job['label'])
    try:
        result = _do_run_job(job)
    except:
        status['time_finished'] = _make_timestamp()
        status['status'] = 'error'
        status['error'] = 'Exception in _do_run_job'
        val = mt.getValue(key=job)
        if val == 'in-process-' + code:
            _set_job_status(job, status)
        raise

    val = mt.getValue(key=job)
    if val != 'in-process-' + code:
        print('Not saving result because in-process code does not match {} <> {}.'.format(val, 'in-process-' + code))
        return

    status['time_finished'] = _make_timestamp()
    status['result'] = result
    if 'error' in result:
        print('Error running job: ' + result['error'])
        status['status'] = 'error'
        status['error'] = result['error']
        _set_job_status(job, status)
        mt.setValue(key=job, value='error-' + code)
        return
    status['status'] = 'finished'
    mt.saveObject(key=job, object=result)  # Not needed in future, because we should instead use the status object


def assembleBatchResults(*, batch_name):
    batch = mt.loadObject(key=dict(batch_name=batch_name))
    jobs = batch['jobs']

    print('Assembling results for batch {} with {} jobs'.format(batch_name, len(jobs)))
    job_results = []
    for job in jobs:
        print('ASSEMBLING: ' + job['label'])
        result = mt.loadObject(key=job)
        if not result:
            raise Exception('Unable to load object for job: ' + job['label'])
        job_results.append(dict(
            job=job,
            result=result
        ))
    print('Saving results...')
    mt.saveObject(key=dict(name='job_results', batch_name=batch_name), object=dict(job_results=job_results))
    print('Done.')


def _clear_job_results(*, jobs, incomplete_only=True):
    for job in jobs:
        val = mt.getValue(key=job)
        if val:
            if (not incomplete_only) or (val.startswith('in-process')) or (val.startswith('error')):
                print('Clearing job: ' + job['label'])
                mt.setValue(key=job, value=None)


def _download_recordings(*, jobs):
    for _, job in enumerate(jobs):
        val = mt.getValue(key=job)
        if not val:
            if 'recording' in job:
                if 'directory' in job['recording']:
                    dsdir = job['recording']['directory']
                    fname = dsdir + '/raw.mda'
                    print('REALIZING FILE: ' + fname)
                    mt.realizeFile(path=fname)


def _make_timestamp():
    return str(datetime.datetime.now())



import vdomr as vd
import spikeforestwidgets as SFW
from spikeforest import EfficientAccessRecordingExtractor
import spikeextractors as se
import time
import multiprocessing
import sys
from .stdoutsender import StdoutSender
import mtlogging


class TimeseriesView(vd.Component):
    def __init__(self, *, context, opts, prepare_result):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)
        self._opts = opts

        earx = EfficientAccessRecordingExtractor(path=prepare_result['earx_file'])
        self._timeseries_widget = SFW.TimeseriesWidget(recording=earx)
        self._timeseries_widget.setSize(self._size)

    # this will be done in a worker thread
    @staticmethod
    def prepareView(context, opts):
        print('***** Initializing context...')
        context.initialize()

        rx = context.recordingExtractor()

        print('***** Preparing efficient access recording extractor...')
        earx = EfficientAccessRecordingExtractor(recording=rx)

        print('***** Precomputing multiscale recordings...')
        SFW.precomputeMultiscaleRecordings(recording=earx)
        return dict(
            earx_file=earx.path()
        )

    def setSize(self, size):
        self._size = size
        if self._timeseries_widget:
            self._timeseries_widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Timeseries'

    def title(self):
        return 'Timeseries for {}'.format(self._context.recordingLabel())

    def render(self):
        return vd.div(
            self._timeseries_widget
        )



import numpy as np
from .synthesize_single_waveform import synthesize_single_waveform


def synthesize_random_waveforms(*, M=5, T=500, K=20, upsamplefac=13, timeshift_factor=0.3, average_peak_amplitude=-10):
    geometry = None
    avg_durations = [200, 10, 30, 200]
    avg_amps = [0.5, 10, -1, 0]
    rand_durations_stdev = [10, 4, 6, 20]
    rand_amps_stdev = [0.2, 3, 0.5, 0]
    rand_amp_factor_range = [0.5, 1]
    # geom_spread_coef1 = 0.2
    # geom_spread_coef2 = 1
    geom_spread_coef1 = 0.5
    geom_spread_coef2 = 0.4

    if not geometry:
        geometry = np.zeros((2, M))
        geometry[0, :] = np.arange(1, M + 1)

    geometry = np.array(geometry)
    avg_durations = np.array(avg_durations)
    avg_amps = np.array(avg_amps)
    rand_durations_stdev = np.array(rand_durations_stdev)
    rand_amps_stdev = np.array(rand_amps_stdev)
    rand_amp_factor_range = np.array(rand_amp_factor_range)

    neuron_locations = get_default_neuron_locations(M, K, geometry)

    # The waveforms_out
    WW = np.zeros((M, T * upsamplefac, K))

    for k in range(1, K + 1):
        for m in range(1, M + 1):
            diff = neuron_locations[:, k - 1] - geometry[:, m - 1]
            dist = np.sqrt(np.sum(diff ** 2))
            durations0 = np.maximum(np.ones(avg_durations.shape),
                                    avg_durations + np.random.randn(1, 4) * rand_durations_stdev) * upsamplefac
            amps0 = avg_amps + np.random.randn(1, 4) * rand_amps_stdev
            waveform0 = synthesize_single_waveform(N=T * upsamplefac, durations=durations0, amps=amps0)
            waveform0 = np.roll(waveform0, int(timeshift_factor * dist * upsamplefac))
            waveform0 = waveform0 * np.random.uniform(rand_amp_factor_range[0], rand_amp_factor_range[1])
            WW[m - 1, :, k - 1] = waveform0 / (geom_spread_coef1 + dist * geom_spread_coef2)

    peaks = np.max(np.abs(WW), axis=(0, 1))
    WW = WW / np.mean(peaks) * average_peak_amplitude

    return (WW, geometry.T)


def get_default_neuron_locations(M, K, geometry):
    num_dims = geometry.shape[0]
    neuron_locations = np.zeros((num_dims, K))
    for k in range(1, K + 1):
        ind = k / (K + 1) * M
        ind0 = int(ind)
        p = ind - ind0
        neuron_locations[:, k - 1] = (1 - p) * geometry[:, ind0 - 1] + p * geometry[:, ind0]
    return neuron_locations



import vdomr as vd


class TestView(vd.Component):
    def __init__(self, context, opts=None, prepare_result=None):
        vd.Component.__init__(self)
        self._context = context
        self._opts = opts
        self._size = (100, 100)
        self._widget = TestWidget(context)

    @staticmethod
    def prepareView(context, opts):
        # prepare code goes here
        # Or, you can remove this function altogether
        pass

    def setSize(self, size):
        self._size = size
        self._widget.setSize(size)

    def size(self):
        return self._size

    def tabLabel(self):
        return 'Test view'

    def render(self):
        return self._widget


class TestWidget(vd.Component):
    def __init__(self, context):
        vd.Component.__init__(self)
        self._context = context
        self._size = (100, 100)

    def setSize(self, size):
        self._size = size

    def size(self):
        return self._size

    def render(self):
        return vd.div('Test created via snippet.')



from mountaintools import client as mt
import spikeextractors as si
import mlprocessors as mlpr
import os
import shutil
import random
import string
import multiprocessing
import mtlogging
# from . import sorters as sorters

from spikesorters import MountainSort4, SpykingCircus, YASS, YASS1, IronClust, KiloSort, KiloSort2, MountainSort4TestError, HerdingSpikes2, JRClust, Klusta, Tridesclous, Waveclus

Processors = dict(
    MountainSort4=(MountainSort4, 'default'),
    IronClust=(IronClust, None),
    SpykingCircus=(SpykingCircus, 'default'),
    KiloSort=(KiloSort, None),
    KiloSort2=(KiloSort2, None),
    Yass=(YASS, 'default'),
    Yass1=(YASS1, 'default'),
    MountainSort4TestError=(MountainSort4TestError, 'default'),
    HerdingSpikes2=(HerdingSpikes2, 'default'),
    JRClust=(JRClust, None),
    Klusta=(Klusta, None),
    Tridesclous=(Tridesclous, None),
    Waveclus=(Waveclus, None),
)


def find_sorter_processor_and_container(processor_name):
    if processor_name not in Processors:
        raise Exception('No such sorter: ' + processor_name)
    SS = Processors[processor_name][0]
    SS_container = Processors[processor_name][1]
    if SS_container:
        if SS_container == 'default':
            SS_container = SS.CONTAINER
        if SS_container:
            print('Locating container: ' + SS_container)
            if not mt.findFile(path=SS_container):
                raise Exception('Unable to realize container: ' + SS_container)
    return SS, SS_container


@mtlogging.log()
def sort_recordings(*, sorter, recordings, num_workers=None, disable_container=False, compute_resource, job_timeout=60 * 20, label=None):
    print('')
    print('>>>>>> {}'.format(label or 'sort recordings'))
    sorting_params = sorter['params']
    processor_name = sorter['processor_name']
    if processor_name in Processors:
        SS = Processors[processor_name][0]
        SS_container = Processors[processor_name][1]
        if disable_container:
            SS_container = None
    else:
        raise Exception('No such sorter: ' + processor_name)

    if SS_container:
        if SS_container == 'default':
            SS_container = SS.CONTAINER
        print('Locating container: ' + SS_container)
        if not mt.findFile(path=SS_container):
            raise Exception('Unable to realize container: ' + SS_container)

    print('>>>>>>>>>>> Sorting recordings using {}'.format(processor_name))

    # pool = multiprocessing.Pool(20)
    # sorting_jobs=pool.map(_create_sorting_job_for_recording_helper, [dict(recording=recording, sorter=sorter, job_timeout=job_timeout) for recording in recordings])
    # pool.close()
    # pool.join()

    sorting_params = sorter['params']
    processor_name = sorter['processor_name']

    sorting_jobs = SS.createJobs([
        dict(
            _container=SS_container,
            _timeout=job_timeout,
            _label='Sort recording {}/{} using {}'.format(recording.get(
                'study', ''), recording.get('name', ''), sorter.get('name', '')),
            _additional_files_to_realize=[recording['directory'] + '/raw.mda'],
            _compute_resource=sorter.get('compute_resource', None),
            recording_dir=recording['directory'],
            channels=recording.get('channels', []),
            firings_out=dict(ext='.mda', upload=True),
            **sorting_params
        )
        for recording in recordings
    ])

    label = label or 'Sort recordings using {}'.format(processor_name)
    sorting_job_results = mlpr.executeBatch(
        jobs=sorting_jobs, label=label, num_workers=num_workers, compute_resource=compute_resource)

    print('Gathering sorting results...')
    sorting_results = [
        dict(
            recording=recording,
            sorter=sorter,
            firings_true=recording['directory'] + '/firings_true.mda',
            processor_name=SS.NAME,
            processor_version=SS.VERSION,
            execution_stats=sorting_job_results[ii].runtime_info,
            console_out=sorting_job_results[ii].console_out,
            container=SS_container,
            firings=(sorting_job_results[ii].outputs or dict()).get(
                'firings_out', None)
        )
        for ii, recording in enumerate(recordings)
    ]

    return sorting_results


@mtlogging.log()
def multi_sort_recordings(*, sorters, recordings, num_workers=None, disable_container=False, job_timeout=60 * 20, label=None, upload_to=None, skip_failing=None):
    print('')
    print('>>>>>> {}'.format(label or 'sort recordings'))

    sorting_jobs_by_compute_resource = dict()
    sorting_results_by_compute_resource = dict()

    # initialize
    for sorter in sorters:
        compute_resource = sorter.get('compute_resource', None)
        if compute_resource not in sorting_jobs_by_compute_resource:
            sorting_jobs_by_compute_resource[compute_resource] = []
            sorting_results_by_compute_resource[compute_resource] = []

    sorting_results = []
    for sorter in sorters:
        sorting_params = sorter['params']
        processor_name = sorter['processor_name']
        if processor_name in Processors:
            SS = Processors[processor_name][0]
            SS_container = Processors[processor_name][1]
            if disable_container:
                SS_container = None
        else:
            raise Exception('No such sorter: ' + processor_name)

        if hasattr(SS, 'install'):
            SS.install()

        if SS_container:
            if SS_container == 'default':
                SS_container = SS.CONTAINER
            print('Locating container: ' + SS_container)
            if not mt.findFile(path=SS_container):
                raise Exception('Unable to realize container: ' + SS_container)

        sorting_params = sorter['params']
        processor_name = sorter['processor_name']
        compute_resource = sorter.get('compute_resource', None)

        sorting_jobs0 = SS.createJobs([
            dict(
                _container=SS_container,
                _timeout=job_timeout,
                _skip_failing=skip_failing,
                _label='Sort recording {}/{} using {}'.format(recording.get(
                    'study', ''), recording.get('name', ''), sorter.get('name', '')),
                _additional_files_to_realize=[
                    recording['directory'] + '/raw.mda'],
                _compute_resource=sorter.get('compute_resource', None),
                recording_dir=recording['directory'],
                channels=recording.get('channels', []),
                firings_out=dict(ext='.mda', upload=True),
                **sorting_params
            )
            for recording in recordings
        ])

        sorting_jobs_by_compute_resource[compute_resource].extend(
            sorting_jobs0)

        sorting_results_by_compute_resource[compute_resource].extend([
            dict(
                recording=recording,
                sorter=sorter,
                firings_true=recording['directory'] + '/firings_true.mda',
                processor_name=SS.NAME,
                processor_version=SS.VERSION,
                container=SS_container,
                # firings, execution_stats, and console_out will be filled in once the job completes
            )
            for ii, recording in enumerate(recordings)
        ])

    sorting_results = []
    for compute_resource, sorting_jobs0 in sorting_jobs_by_compute_resource.items():
        label = label or 'Sort recordings (resource={})'.format(
            compute_resource)
        sorting_job_results0 = mlpr.executeBatch(
            jobs=sorting_jobs0, label=label, num_workers=num_workers, compute_resource=compute_resource)
        sorting_results0 = sorting_results_by_compute_resource[compute_resource]
        for ii, sr in enumerate(sorting_results0):
            sr['execution_stats'] = sorting_job_results0[ii].runtime_info
            sr['console_out'] = sorting_job_results0[ii].console_out
            sr['firings'] = (sorting_job_results0[ii].outputs or dict()).get(
                'firings_out', None)
            if upload_to:
                mt.createSnapshot(path=sr['console_out'], upload_to=upload_to)
                if sr['firings']:
                    mt.createSnapshot(path=sr['firings'], upload_to=upload_to)
        sorting_results.extend(sorting_results0)

    return sorting_results



from .filterrecording import FilterRecording
import numpy as np


class WhitenRecording(FilterRecording):
    def __init__(self, *, recording):
        FilterRecording.__init__(self, recording=recording, chunk_size=1000)
        self._recording = recording
        self._whitening_matrix = self._compute_whitening_matrix()

    def _get_random_data_for_whitening(self, num_chunks=50, chunk_size=500):
        N = self._recording.get_num_frames()
        list = []
        for i in range(num_chunks):
            ff = np.random.randint(0, N - chunk_size)
            chunk = self._recording.get_traces(start_frame=ff, end_frame=ff + chunk_size)
            list.append(chunk)
        return np.concatenate(list, axis=1)

    def _compute_whitening_matrix(self):
        data = self._get_random_data_for_whitening()
        AAt = data @ np.transpose(data)
        AAt = AAt / data.shape[1]
        U, S, Ut = np.linalg.svd(AAt, full_matrices=True)
        W = (U @ np.diag(1 / np.sqrt(S))) @ Ut
        return W

    def filterChunk(self, *, start_frame, end_frame):
        chunk = self._recording.get_traces(start_frame=start_frame, end_frame=end_frame)
        chunk2 = self._whitening_matrix @ chunk
        return chunk2


def whiten(recording):
    return WhitenRecording(
        recording=recording
    )
